import os

import re
import time

import ffmpeg
import shutil

from sqlalchemy.orm import aliased
from sqlalchemy.orm.exc import ObjectDeletedError

from zou.app import config
from zou.app.stores import file_store

from zou.app.models.entity import Entity
from zou.app.models.preview_file import PreviewFile
from zou.app.models.project import Project
from zou.app.models.project_status import ProjectStatus
from zou.app.models.task import Task
from zou.app.models.task_type import TaskType
from zou.app.services import (
    names_service,
    files_service,
    assets_service,
    shots_service,
    projects_service,
    tasks_service,
)
from zou.utils import movie
from zou.app.utils import (
    events,
    fields,
    remote_job,
    thumbnail as thumbnail_utils,
)
from zou.app.services.exception import (
    WrongParameterException,
    PreviewFileNotFoundException,
    ProjectNotFoundException,
    EpisodeNotFoundException,
)
from zou.app.utils import fs


def get_preview_file_dimensions(project, entity=None):
    """
    Return dimensions set at entity level or project level or default
    dimensions if the dimensions are not set.
    Entity resolution has priority over project resolution.
    The default size is based on 1080 height
    """
    resolution = project["resolution"]
    entity_data = {}
    if entity is not None:
        entity_data = entity.get("data", {}) or {}
    entity_resolution = entity_data.get("resolution", None)
    width = None
    height = 1080

    if _is_valid_resolution(entity_resolution):
        resolution = entity_resolution

    if _is_valid_resolution(resolution):
        [width, height] = resolution.split("x")
        width = int(width)
        height = int(height)

    if _is_valid_partial_resolution(resolution):
        [width, height] = resolution.split("x")
        width = None
        height = int(height)

    return (width, height)


def _is_valid_resolution(resolution):
    """
    Return true if the dimension follows the 1920x1080 pattern.
    """
    return resolution is not None and bool(
        re.match(r"\d{3,4}x\d{3,4}", resolution)
    )


def _is_valid_partial_resolution(resolution):
    """
    Return true if the dimension follows the x1080 pattern.
    """
    return resolution is not None and bool(re.match(r"x\d{3,4}", resolution))


def get_preview_file_fps(project, entity=None):
    """
    Return fps set at project level or default fps if the dimensions are not
    set.
    """
    fps = "25.00"
    if project.get("fps", None) is not None:
        fps = project["fps"].replace(",", ".")

    if entity is not None:
        entity_data = entity.get("data", {}) or {}
        if entity_data.get("fps", None) is not None:
            fps = entity_data["fps"].replace(",", ".")

    return "%.3f" % float(fps)


def get_project_from_preview_file(preview_file_id):
    """
    Get project dict of related preview file.
    """
    preview_file = files_service.get_preview_file_raw(preview_file_id)

    task = Task.get(preview_file.task_id)
    project = Project.get(task.project_id)
    return project.serialize()


def get_entity_from_preview_file(preview_file_id):
    """
    Get entity dict of related preview file.
    """
    preview_file = files_service.get_preview_file_raw(preview_file_id)
    task = Task.get(preview_file.task_id)
    entity = Entity.get(task.entity_id)
    return entity.serialize()


def update_preview_file(preview_file_id, data, silent=False):
    try:
        preview_file = files_service.get_preview_file_raw(preview_file_id)
    except BaseException:
        # Dirty hack because sometimes the preview file retrieval crashes.
        try:
            time.sleep(1)
            preview_file = files_service.get_preview_file_raw(preview_file_id)
        except BaseException:
            time.sleep(5)
            preview_file = files_service.get_preview_file_raw(preview_file_id)
    return update_preview_file_raw(preview_file, data, silent=silent)


def update_preview_file_raw(preview_file, data, silent=False):
    preview_file.update(data)
    files_service.clear_preview_file_cache(str(preview_file.id))
    if not silent:
        task = Task.get(preview_file.task_id)
        events.emit(
            "preview-file:update",
            {"preview_file_id": str(preview_file.id)},
            project_id=str(task.project_id),
        )
    return preview_file.serialize()


def set_preview_file_as_broken(preview_file_id):
    """
    Mark given preview file as broken.
    """
    return update_preview_file(preview_file_id, {"status": "broken"})


def set_preview_file_as_ready(preview_file_id):
    """
    Mark given preview file as ready.
    """
    return update_preview_file(preview_file_id, {"status": "ready"})


def prepare_and_store_movie(
    preview_file_id,
    uploaded_movie_path,
    normalize=True,
    add_source_to_file_store=True,
):
    """
    Prepare movie preview, normalize the movie as a .mp4, build the thumbnails
    and store the files.
    """
    from zou.app import app as current_app

    with current_app.app_context():
        preview_file_raw = files_service.get_preview_file_raw(preview_file_id)
        normalized_movie_low_path = None
        try:
            project = get_project_from_preview_file(preview_file_id)
            entity = get_entity_from_preview_file(preview_file_id)
        except PreviewFileNotFoundException:
            time.sleep(2)
            try:
                project = get_project_from_preview_file(preview_file_id)
                entity = get_entity_from_preview_file(preview_file_id)
            except PreviewFileNotFoundException:
                current_app.logger.error(
                    "Data is missing from database", exc_info=1
                )
                time.sleep(2)
                preview_file = set_preview_file_as_broken(preview_file_id)
                return preview_file

        fps = get_preview_file_fps(project, entity)
        (width, height) = get_preview_file_dimensions(project, entity)

        if normalize:
            current_app.logger.info("start normalization")
            try:
                if add_source_to_file_store:
                    file_store.add_movie(
                        "source", preview_file_id, uploaded_movie_path
                    )
                if (
                    config.ENABLE_JOB_QUEUE_REMOTE
                    and len(config.JOB_QUEUE_NOMAD_NORMALIZE_JOB) > 0
                ):
                    result = _run_remote_normalize_movie(
                        current_app, preview_file_id, fps, width, height
                    )
                    if result is True:
                        err = None
                    else:
                        err = result

                    normalized_movie_path = fs.get_file_path_and_file(
                        config,
                        file_store.get_local_movie_path,
                        file_store.open_movie,
                        "previews",
                        preview_file_id,
                        ".mp4",
                    )
                else:
                    (
                        normalized_movie_path,
                        normalized_movie_low_path,
                        err,
                    ) = movie.normalize_movie(
                        uploaded_movie_path,
                        fps=fps,
                        width=width,
                        height=height,
                    )
                    file_store.add_movie(
                        "previews", preview_file_id, normalized_movie_path
                    )
                    file_store.add_movie(
                        "lowdef", preview_file_id, normalized_movie_low_path
                    )
                if err:
                    current_app.logger.error(
                        "Fail to normalize: %s" % uploaded_movie_path
                    )
                    current_app.logger.error(err)

                current_app.logger.info(
                    "file normalized %s" % normalized_movie_path
                )
                current_app.logger.info("file stored")
            except Exception as exc:
                if isinstance(exc, ffmpeg.Error):
                    current_app.logger.error(exc.stderr)
                current_app.logger.error("failed", exc_info=1)
                preview_file = set_preview_file_as_broken(preview_file_id)
                return preview_file
        else:
            file_store.add_movie(
                "previews", preview_file_id, uploaded_movie_path
            )
            file_store.add_movie(
                "lowdef", preview_file_id, uploaded_movie_path
            )
            normalized_movie_path = uploaded_movie_path

        # Build thumbnails
        size = movie.get_movie_size(normalized_movie_path)
        width, height = size
        original_picture_path = movie.generate_thumbnail(normalized_movie_path)
        thumbnail_utils.turn_into_thumbnail(original_picture_path, size)
        save_variants(preview_file_id, original_picture_path)
        file_size = os.path.getsize(normalized_movie_path)
        duration = movie.get_movie_duration(normalized_movie_path)
        current_app.logger.info("thumbnail created %s" % original_picture_path)

        # Build tiles
        try:
            tile_path = movie.generate_tile(normalized_movie_path)
            file_store.add_picture("tiles", preview_file_id, tile_path)
            os.remove(tile_path)
            current_app.logger.info("tile created %s" % tile_path)
        except Exception:
            current_app.logger.error("Failed to create tile", exc_info=1)

        # Remove files and update status
        os.remove(uploaded_movie_path)
        if normalize:
            os.remove(normalized_movie_path)
            if normalized_movie_low_path:
                os.remove(normalized_movie_low_path)

        # Save metadata, save preview file id in the related task and entity
        preview_file = update_preview_file_raw(
            preview_file_raw,
            {
                "status": "ready",
                "file_size": file_size,
                "width": width,
                "height": height,
                "duration": duration,
            },
        )
        tasks_service.update_preview_file_info(preview_file)
        return preview_file


def _run_remote_normalize_movie(app, preview_file_id, fps, width, height):
    params = {
        "version": "1",
        "preview_file_id": preview_file_id,
        "width": width,
        "height": height,
        "fps": fps,
    }
    nomad_job = app.config["JOB_QUEUE_NOMAD_NORMALIZE_JOB"]
    result = remote_job.run_job(app, config, nomad_job, params)
    return result


def save_variants(preview_file_id, original_picture_path, with_original=True):
    """
    Build variants of a picture file and save them in the main storage.
    """
    variants = thumbnail_utils.generate_preview_variants(
        original_picture_path, preview_file_id
    )
    if with_original:
        variants.append(("original", original_picture_path))
    for prefix, path in variants:
        file_store.add_picture(prefix, preview_file_id, path)
        os.remove(path)
        clear_variant_from_cache(preview_file_id, prefix)

    return variants


def clear_variant_from_cache(preview_file_id, prefix, extension="png"):
    """
    Clear a variant from the cache to force to redownload from object storage.
    """
    if config.FS_BACKEND != "local":
        file_path = os.path.join(
            config.TMP_DIR,
            "cache-%s-%s.%s" % (prefix, preview_file_id, extension),
        )
        if os.path.exists(file_path):
            os.remove(file_path)
    return preview_file_id


def update_preview_file_position(preview_file_id, position):
    """
    Change positions for preview files of given task and revision.
    Given position is the new position for given preview file.
    """
    preview_file = files_service.get_preview_file_raw(preview_file_id)
    task_id = preview_file.task_id
    revision = preview_file.revision
    preview_files = (
        PreviewFile.query.filter_by(task_id=task_id, revision=revision)
        .order_by(PreviewFile.position, PreviewFile.created_at)
        .all()
    )
    if position > 0 and position <= len(preview_files):
        tmp_list = [p for p in preview_files if str(p.id) != preview_file_id]
        tmp_list.insert(position - 1, preview_file)
        for i, preview in enumerate(tmp_list):
            preview.update({"position": i + 1})
    return PreviewFile.serialize_list(preview_files)


def get_preview_files_for_revision(task_id, revision):
    """
    Get all preview files for given task and revision.
    """
    preview_files = PreviewFile.query.filter_by(
        task_id=task_id, revision=revision
    ).order_by(PreviewFile.position)
    return fields.serialize_models(preview_files)


def update_preview_file_annotations(
    person_id,
    project_id,
    preview_file_id,
    additions=[],
    updates=[],
    deletions=[],
):
    """
    Update annotations for given preview file.
    """
    preview_file = files_service.get_preview_file_raw(preview_file_id)
    previous_annotations = preview_file.annotations or []
    annotations = _clean_annotations(previous_annotations)
    annotations = _apply_annotation_additions(previous_annotations, additions)
    annotations = _apply_annotation_updates(annotations, updates)
    annotations = _apply_annotation_deletions(annotations, deletions)
    preview_file.update({"annotations": []})
    preview_file.update({"annotations": annotations})
    files_service.clear_preview_file_cache(preview_file_id)
    preview_file = files_service.get_preview_file(preview_file_id)
    events.emit(
        "preview-file:annotation-update",
        {
            "preview_file_id": preview_file_id,
            "person_id": person_id,
            "updated_at": preview_file["updated_at"],
        },
        project_id=project_id,
    )
    return preview_file


def _clean_annotations(annotations):
    for annotation in annotations:
        objects = annotation.get("drawing", {}).get("objects", [])
        for current_object in objects:
            if (
                "id" not in current_object
                or len(current_object["id"]) == 0
                or current_object["id"] is None
            ):
                current_object["id"] = str(fields.gen_uuid())
    return annotations


def _apply_annotation_additions(previous_annotations, new_annotations):
    annotations = list(previous_annotations)
    annotation_map = _get_annotation_time_map(annotations)

    for new_annotation in new_annotations:
        previous_annotation = annotation_map.get(new_annotation["time"], None)
        if previous_annotation is None:
            new_objects = new_annotation.get("drawing", {}).get("objects", [])
            for new_object in new_objects:
                if (
                    "id" not in new_object
                    or len(new_object["id"]) == 0
                    or new_object["id"] is None
                ):
                    new_object["id"] = str(fields.gen_uuid())
            annotations.append(new_annotation)
        else:
            previous_objects = previous_annotation.get("drawing", {}).get(
                "objects", []
            )
            new_objects = new_annotation.get("drawing", {}).get("objects", [])
            for new_object in new_objects:
                if "id" not in new_object or len(new_object["id"]) == 0:
                    new_object["id"] = str(fields.gen_uuid())
            previous_annotation["drawing"]["objects"] = _get_new_annotations(
                previous_objects, new_objects
            )
    return annotations


def _get_new_annotations(previous_objects, new_objects):
    result = list(previous_objects)
    previous_map = {}
    for previous_object in result:
        if "id" not in previous_object or len(previous_object["id"]) == 0:
            previous_object["id"] = str(fields.gen_uuid())
        previous_map[previous_object["id"]] = True

    for new_object in new_objects:
        object_id = new_object.get("id", "")
        if object_id not in previous_map:
            result.append(new_object)
    return result


def _apply_annotation_updates(annotations, updates):
    annotation_map = _get_annotation_time_map(annotations)
    for update in updates:
        time = update["time"]
        if time in annotation_map:
            result = []
            previous_object_map = {}
            update_map = {}
            annotation = annotation_map[time]

            previous_objects = annotation.get("drawing", {}).get("objects", [])
            for previous_object in previous_objects:
                if "id" in previous_object:
                    previous_object_map[previous_object["id"]] = (
                        previous_object
                    )

            updated_objects = update.get("drawing", {}).get("objects", [])
            for updated_object in updated_objects:
                if "id" in updated_object:
                    update_map[updated_object["id"]] = update

            result = [
                previous_object
                for previous_object in previous_objects
                if previous_object.get("id", None) not in update_map
            ]
            for updated_object in updated_objects:
                if (
                    "id" in updated_object
                    and updated_object["id"] in previous_object_map
                ):
                    result.append(updated_object)
            annotation["drawing"]["objects"] = result
    return annotations


def _apply_annotation_deletions(annotations, deletions):
    annotation_map = _get_annotation_time_map(annotations)

    for deletion in deletions:
        if deletion["time"] in annotation_map:
            annotation = annotation_map[deletion["time"]]
            deleted_object_ids = deletion.get("objects", [])
            previous_objects = annotation.get("drawing", {}).get("objects", [])
            annotation.get("drawing", {})["objects"] = [
                previous_object
                for previous_object in previous_objects
                if previous_object.get("id", "") not in deleted_object_ids
            ]

    return _clear_empty_annotations(annotations)


def _get_annotation_time_map(annotations):
    annotation_map = {}
    for annotation in annotations:
        annotation_map[annotation["time"]] = annotation
    return annotation_map


def _clear_empty_annotations(annotations):
    return [
        annotation
        for annotation in annotations
        if len(annotation.get("drawing", {}).get("objects", [])) > 0
    ]


def get_running_preview_files():
    """
    Return preview files for all productions with status equals to broken
    or processing.
    """
    entries = (
        PreviewFile.query.join(Task)
        .join(Project)
        .join(ProjectStatus)
        .filter(ProjectStatus.name.in_(("Active", "open", "Open")))
        .filter(PreviewFile.status.in_(("broken", "processing")))
        .add_columns(Task.project_id, Task.task_type_id, Task.entity_id)
        .order_by(PreviewFile.created_at.desc())
    )

    results = []
    for preview_file, project_id, task_type_id, entity_id in entries:
        result = preview_file.serialize()
        result["project_id"] = fields.serialize_value(project_id)
        result["task_type_id"] = fields.serialize_value(task_type_id)
        result["full_entity_name"], _, _ = names_service.get_full_entity_name(
            entity_id
        )
        results.append(result)
    return results


def get_preview_files_for_entity(entity_id):
    """
    Return all preview files related to given entity.
    """
    query = (
        PreviewFile.query.join(Task)
        .join(TaskType)
        .filter(Task.entity_id == entity_id)
        .order_by(
            TaskType.name, PreviewFile.revision.desc(), PreviewFile.position
        )
    )
    return [preview_file.present() for preview_file in query.all()]


def get_last_preview_file_for_task(task_id):
    """
    Get last preview published for given task.
    """
    preview = (
        PreviewFile.query.filter(PreviewFile.task_id == task_id)
        .order_by(
            PreviewFile.revision.desc(),
            PreviewFile.created_at,
        )
        .first()
    )
    if preview is None:
        return None
    else:
        return preview.serialize()


def extract_frame_from_preview_file(preview_file, frame_number):
    try:
        project = get_project_from_preview_file(preview_file["id"])
    except PreviewFileNotFoundException:
        raise PreviewFileNotFoundException

    if preview_file["extension"] == "mp4":
        preview_file_path = fs.get_file_path_and_file(
            config,
            file_store.get_local_movie_path,
            file_store.open_movie,
            "previews",
            preview_file["id"],
            "mp4",
        )
    else:
        raise PreviewFileNotFoundException

    fps = get_preview_file_fps(
        project, get_entity_from_preview_file(preview_file["id"])
    )
    extracted_frame_path = movie.extract_frame_from_movie(
        preview_file_path, frame_number, fps
    )

    return extracted_frame_path


def replace_extracted_frame_for_preview_file(preview_file, frame_number):
    extracted_frame_path = extract_frame_from_preview_file(
        preview_file, frame_number
    )
    extracted_frame_path = thumbnail_utils.turn_into_thumbnail(
        extracted_frame_path
    )
    save_variants(preview_file["id"], extracted_frame_path)


def extract_tile_from_preview_file(preview_file):
    if preview_file["extension"] == "mp4":
        preview_file_path = fs.get_file_path_and_file(
            config,
            file_store.get_local_movie_path,
            file_store.open_movie,
            "previews",
            preview_file["id"],
            "mp4",
        )
        extracted_tile_path = movie.generate_tile(preview_file_path)
        return extracted_tile_path
    else:
        raise WrongParameterException("Preview file is not a movie")


def reset_movie_files_metadata():
    """
    Reset preview files size informations of open projects.
    """
    preview_files = (
        PreviewFile.query.join(Task)
        .join(Project)
        .join(ProjectStatus)
        .filter(ProjectStatus.name.in_(("Active", "open", "Open")))
        .filter(PreviewFile.status.not_in(("broken", "processing")))
        .filter(PreviewFile.extension == "mp4")
    )
    for preview_file in preview_files:
        try:
            preview_file_path = fs.get_file_path_and_file(
                config,
                file_store.get_local_movie_path,
                file_store.open_movie,
                "previews",
                str(preview_file.id),
                "mp4",
            )
            file_size = os.path.getsize(preview_file_path)
            width, height = movie.get_movie_size(preview_file_path)
            duration = float(movie.get_movie_duration(preview_file_path))
            update_preview_file_raw(
                preview_file,
                {
                    "width": width,
                    "height": height,
                    "file_size": file_size,
                    "duration": duration,
                },
            )
            print(
                f"Size information stored preview file {preview_file.id}",
            )
        except Exception as e:
            print(
                f"Failed to store information for preview file {preview_file.id}: {e}"
            )


def reset_picture_files_metadata():
    """
    Reset preview files size informations of open projects.
    """
    preview_files = (
        PreviewFile.query.join(Task)
        .join(Project)
        .join(ProjectStatus)
        .filter(ProjectStatus.name.in_(("Active", "open", "Open")))
        .filter(PreviewFile.status.not_in(("broken", "processing")))
        .filter(PreviewFile.extension == "png")
    )
    for preview_file in preview_files:
        try:
            preview_file_path = fs.get_file_path_and_file(
                config,
                file_store.get_local_picture_path,
                file_store.open_picture,
                "original",
                str(preview_file.id),
                "png",
            )
            width, height = thumbnail_utils.get_dimensions(preview_file_path)
            file_size = os.path.getsize(preview_file_path)
            update_preview_file_raw(
                preview_file,
                {
                    "width": width,
                    "height": height,
                    "file_size": file_size,
                },
            )
            print(
                f"Size information stored for preview file {preview_file.id}",
            )
        except Exception as e:
            print(
                f"Failed to store information for preview file {preview_file.id}: {e}"
            )


def generate_preview_extra(
    project=None,
    entity_id=None,
    episodes=[],
    only_shots=False,
    only_assets=False,
    force_regenerate_tiles=False,
    with_tiles=False,
    with_metadata=False,
    with_thumbnails=False,
):
    """
    Generate tiles for all movie previews and reset previews file size
    informations of open projects.
    """
    print("Generating preview extras...")
    query = (
        PreviewFile.query.join(Task)
        .join(Entity)
        .join(Project)
        .join(ProjectStatus)
        .filter(ProjectStatus.name.in_(("Active", "open", "Open")))
        .filter(PreviewFile.status.not_in(("broken", "processing")))
        .filter(PreviewFile.extension.in_(("mp4", "png")))
    )
    if project is not None:
        try:
            project_id = projects_service.get_project_by_name(project)["id"]
        except ProjectNotFoundException:
            project_id = projects_service.get_project(project)["id"]
        query = query.filter(Project.id == project_id)

    if entity_id is not None:
        query = query.filter(Task.entity_id == entity_id)

    if episodes:
        get_episode_by_name = False
        if project is not None:
            get_episode_by_name = True
        episode_ids = []
        for episode in episodes:
            episode_id = None
            try:
                episode_id = shots_service.get_episode(episode)["id"]
            except EpisodeNotFoundException as e:
                if get_episode_by_name:
                    episode_id = shots_service.get_episode_by_name(
                        project_id, episode
                    )["id"]
                else:
                    raise e
            episode_ids.append(episode_id)
        Sequence = aliased(Entity)
        query = query.join(Sequence, Sequence.id == Entity.parent_id).filter(
            Sequence.parent_id.in_(episode_ids)
        )
    if only_shots:
        query = query.filter(
            Entity.entity_type_id == shots_service.get_shot_type()["id"]
        )
    elif only_assets:
        query = query.filter(
            Entity.entity_type_id.not_in(
                assets_service.get_temporal_type_ids()
            )
        )

    total = query.count()
    print(f"{total} previews found.")
    for index, preview_file in enumerate(query.all()):
        try:
            preview_file_id = str(preview_file.id)
        except ObjectDeletedError:
            continue
        prefix = "previews" if preview_file.extension == "mp4" else "original"
        if config.FS_BACKEND != "local":
            preview_file_already_in_cache = os.path.isfile(
                os.path.join(
                    config.TMP_DIR,
                    "cache-%s-%s.%s"
                    % (prefix, preview_file_id, preview_file.extension),
                )
            )
        try:
            preview_file_path = _retrieve_preview_file(
                config, file_store, prefix, preview_file
            )
            if with_tiles:
                _generate_tiles(
                    file_store,
                    preview_file,
                    preview_file_path,
                    total,
                    index + 1,
                    force=force_regenerate_tiles,
                )
            if with_metadata:
                _reset_preview_file_metadata(
                    preview_file, preview_file_path, total, index + 1
                )
            if with_thumbnails:
                _generate_thumbnails(
                    preview_file, preview_file_path, total, index + 1
                )
        finally:
            if (
                config.FS_BACKEND != "local"
                and not preview_file_already_in_cache
            ):
                try:
                    os.remove(preview_file_path)
                except:
                    pass

    print("Extra information generated.")
    return total


def _retrieve_preview_file(config, file_store, prefix, preview_file):
    try:
        preview_file_path = fs.get_file_path_and_file(
            config,
            (
                file_store.get_local_movie_path
                if preview_file.extension == "mp4"
                else file_store.get_local_picture_path
            ),
            (
                file_store.open_movie
                if preview_file.extension == "mp4"
                else file_store.open_picture
            ),
            prefix,
            str(preview_file.id),
            preview_file.extension,
        )
    except Exception as e:
        print(f"Failed to get preview file {preview_file.id}: {e}.")
        return None
    return preview_file_path


def _generate_thumbnails(preview_file, preview_file_path, total, index):
    try:
        original_picture_path = preview_file_path
        if preview_file.extension == "mp4":
            original_picture_path = movie.generate_thumbnail(preview_file_path)
        save_variants(
            preview_file.id, original_picture_path, with_original=False
        )
        print(
            f"{index:0{len(str(total))}}/{total} Thumbnails generated for {preview_file.id}.",
        )
    except Exception as e:
        print(f"Failed to generate thumbnails for {preview_file.id}: {e}.")


def _generate_tiles(
    file_store, preview_file, preview_file_path, total, index, force=False
):
    try:
        if preview_file.extension == "mp4" and (
            force
            or not file_store.exists_picture("tiles", str(preview_file.id))
        ):
            tile_path = movie.generate_tile(preview_file_path)
            file_store.add_picture("tiles", preview_file.id, tile_path)
            os.remove(tile_path)
            print(
                f"{index:0{len(str(total))}}/{total} Tile "
                + f"generated for {preview_file.id}.",
            )
    except Exception as e:
        print(
            f"Failed to generate tile for preview file {preview_file.id}: {e}."
        )


def _reset_preview_file_metadata(
    preview_file, preview_file_path, total, index
):
    try:
        if preview_file.extension == "mp4":
            width, height = movie.get_movie_size(preview_file_path)
        else:
            width, height = thumbnail_utils.get_dimensions(preview_file_path)
        file_size = os.path.getsize(preview_file_path)
        duration = (
            float(movie.get_movie_duration(preview_file_path))
            if preview_file.extension == "mp4"
            else None
        )
        update_preview_file_raw(
            preview_file,
            {
                "width": width,
                "height": height,
                "file_size": file_size,
                "duration": duration,
            },
        )
        print(
            f"{index:0{len(str(total))}}/{total} Size information stored for {preview_file.id}.",
        )
    except Exception as e:
        print(
            f"Failed to store information for preview file {preview_file.id}: {e}.",
        )


def copy_preview_file_on_storage(
    get_path_func,
    exists_func,
    copy_func,
    prefix,
    original_preview_file_id,
    preview_file_to_update_id,
):
    if config.FS_BACKEND == "local":
        file_path = get_path_func(prefix, original_preview_file_id)
        other_file_path = get_path_func(prefix, preview_file_to_update_id)
        if os.path.exists(file_path):
            os.makedirs(os.path.dirname(other_file_path), exist_ok=True)
            shutil.copyfile(file_path, other_file_path)
    elif exists_func(prefix, original_preview_file_id):
        copy_func(
            prefix, original_preview_file_id, prefix, preview_file_to_update_id
        )


def copy_preview_file_in_another_one(
    original_preview_file_id, preview_file_to_update_id
):
    """
    Copy preview file data/files from one preview file to another one.
    """
    original_preview_file = files_service.get_preview_file(
        original_preview_file_id
    )
    is_movie = original_preview_file["extension"] == "mp4"
    is_picture = original_preview_file["extension"] == "png"

    if is_movie:
        prefixes = ["previews", "lowdef"]
        for prefix in prefixes:
            copy_preview_file_on_storage(
                file_store.get_local_movie_path,
                file_store.exists_movie,
                file_store.copy_movie,
                prefix,
                original_preview_file_id,
                preview_file_to_update_id,
            )

    if is_movie or is_picture:
        prefixes = [
            "previews",
            "original",
            "thumbnails",
            "thumbnails-square",
        ]
        if is_movie:
            prefixes.append("tiles")

        for prefix in prefixes:
            copy_preview_file_on_storage(
                file_store.get_local_picture_path,
                file_store.exists_picture,
                file_store.copy_picture,
                prefix,
                original_preview_file_id,
                preview_file_to_update_id,
            )
    else:
        copy_preview_file_on_storage(
            file_store.get_local_file_path,
            file_store.exists_file,
            file_store.copy_file,
            "previews",
            original_preview_file_id,
            preview_file_to_update_id,
        )

    preview_file_to_update = update_preview_file(
        preview_file_to_update_id,
        {
            "extension": original_preview_file["extension"],
            "original_name": original_preview_file["original_name"],
            "status": original_preview_file["status"],
            "file_size": original_preview_file["file_size"],
            "width": original_preview_file["width"],
            "height": original_preview_file["height"],
            "duration": original_preview_file["duration"],
        },
    )
    tasks_service.update_preview_file_info(preview_file_to_update)
    comment = tasks_service.get_comment_by_preview_file_id(
        preview_file_to_update_id
    )
    task = tasks_service.get_task(preview_file_to_update["task_id"])
    comment_id = None
    if comment is not None:
        comment_id = comment["id"]
        events.emit(
            "comment:update",
            {"comment_id": comment_id},
            project_id=task["project_id"],
        )
        events.emit(
            "preview-file:add-file",
            {
                "comment_id": comment_id,
                "task_id": preview_file_to_update["task_id"],
                "preview_file_id": preview_file_to_update["id"],
                "revision": preview_file_to_update["revision"],
                "extension": preview_file_to_update["extension"],
                "status": preview_file_to_update["status"],
            },
            project_id=task["project_id"],
        )

    return preview_file_to_update



"""Change file_size to big integer

Revision ID: df9f8a147e80
Revises: 92bdfe07e5f5
Create Date: 2022-03-14 18:01:49.812243

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "df9f8a147e80"
down_revision = "92bdfe07e5f5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "preview_file",
        "file_size",
        existing_type=sa.INTEGER(),
        type_=sa.BigInteger(),
        existing_nullable=True,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "preview_file",
        "file_size",
        existing_type=sa.BigInteger(),
        type_=sa.INTEGER(),
        existing_nullable=True,
    )
    # ### end Alembic commands ###



"""Person: partial index for email only when not is_bot

Revision ID: 96f58a4a2a58
Revises: b8ed0fb263f8
Create Date: 2024-02-12 03:24:43.601328

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "96f58a4a2a58"
down_revision = "b8ed0fb263f8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.drop_constraint("person_email_key", type_="unique")
        batch_op.create_index(
            "only_one_email_by_person",
            ["email", "is_bot"],
            unique=True,
            postgresql_where=sa.text("is_bot IS NOT true"),
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.drop_index(
            "only_one_email_by_person",
            postgresql_where=sa.text("is_bot IS NOT true"),
        )
        batch_op.create_unique_constraint("person_email_key", ["email"])

    # ### end Alembic commands ###



"""allow attachment on messages

Revision ID: 92b40d79ad3f
Revises: 23122f290ca2
Create Date: 2024-03-08 01:35:16.127135

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "92b40d79ad3f"
down_revision = "23122f290ca2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("chat", schema=None) as batch_op:
        batch_op.drop_index("ix_chat_person_id")
        batch_op.drop_constraint("chat_person_id_fkey", type_="foreignkey")
        batch_op.drop_column("person_id")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    """
    with op.batch_alter_table('chat', schema=None) as batch_op:
        batch_op.add_column(sa.Column('person_id', sa.UUID(), autoincrement=False, nullable=False))
        batch_op.create_foreign_key('chat_person_id_fkey', 'person', ['person_id'], ['id'])
        batch_op.create_index('ix_chat_person_id', ['person_id'], unique=False)
    """

    # ### end Alembic commands ###



"""discord_integration

Revision ID: 92bdfe07e5f5
Revises: aa0a60033106
Create Date: 2022-03-03 17:04:48.006864

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "92bdfe07e5f5"
down_revision = "aa0a60033106"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "organisation",
        sa.Column("chat_token_discord", sa.String(length=80), nullable=True),
    )
    op.add_column(
        "person",
        sa.Column(
            "notifications_discord_enabled", sa.Boolean(), nullable=True
        ),
    )
    op.add_column(
        "person",
        sa.Column(
            "notifications_discord_userid", sa.String(length=60), nullable=True
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("person", "notifications_discord_userid")
    op.drop_column("person", "notifications_discord_enabled")
    op.drop_column("organisation", "chat_token_discord")
    # ### end Alembic commands ###



import meilisearch
from meilisearch.errors import MeilisearchApiError

from zou.app import config
from zou.app.utils import cache

client = None


def init():
    """
    Configure Meilisearch client.
    """
    global client
    if client is None:
        protocol = config.INDEXER["protocol"]
        host = config.INDEXER["host"]
        port = config.INDEXER["port"]
        client = meilisearch.Client(
            f"{protocol}://{host}:{port}",
            config.INDEXER["key"],
            timeout=config.INDEXER["timeout"],
        )


def create_index(index_name, searchable_fields=[], filterable_fields=[]):
    """
    Create a new index and configure it properly by setting searchable_fields
    and allowing to filter on project ids.
    """
    index = None
    try:
        index = get_index(index_name)
    except MeilisearchApiError:
        pass
    if index is None:
        task = client.create_index(index_name, {"primaryKey": "id"})
        client.wait_for_task(
            task.task_uid, timeout_in_ms=config.INDEXER["timeout"]
        )
        index = get_index(index_name)
    index.update_searchable_attributes(searchable_fields)
    index.update_settings(
        {
            "filterableAttributes": filterable_fields,
            "sortableAttributes": [
                "name",
            ],
        }
    )

    return index


@cache.memoize_function(120)
def get_index(index_name):
    """
    Get index matching given name.
    """
    init()
    index = client.get_index(index_name)
    return index


def clear_index(index_name):
    """
    Clear all data into the index matching given name.
    """
    cache.cache.delete_memoized(get_index)
    index = get_index(index_name)
    task = index.delete_all_documents()
    client.wait_for_task(
        task.task_uid, timeout_in_ms=config.INDEXER["timeout"]
    )
    return index


def index_document(index, document):
    """
    Add given document to given index.
    """
    task = index.add_documents([document])
    client.wait_for_task(
        task.task_uid, timeout_in_ms=config.INDEXER["timeout"]
    )
    return index


def index_documents(index, documents):
    """
    Add given documents to given index.
    """
    task = index.add_documents(documents)
    client.wait_for_task(
        task.task_uid, timeout_in_ms=config.INDEXER["timeout"]
    )
    return documents


def search(ix, query, project_ids=[], limit=10, offset=0):
    """
    Perform a search on given index and filter result based on project IDs.
    The number of results can be specified.
    """
    project_ids = ",".join(project_ids)
    search_options = {
        "limit": limit,
        "offset": offset,
        "sort": ["name:asc"],
        "showMatchesPosition": True,
    }
    if len(project_ids) > 0:
        search_options["filter"] = f"project_id IN [{project_ids}]"
    search_results = ix.search(query, search_options)

    results = []
    for result in search_results["hits"]:
        document_id = result["id"]
        matched_fields = []
        for field in result["_matchesPosition"].keys():
            if "metadatas" in field:
                matched_fields.append(field[10:])
            else:
                matched_fields.append(field)
        results.append((document_id, matched_fields))

    return results


def remove_document(index, entry_id):
    """
    Remove document matching given id from given index.
    """
    index.delete_document(entry_id)
    return entry_id



"""Add ready for column to entities

Revision ID: 87efceb6745b
Revises: 398150912a3f
Create Date: 2021-11-09 00:13:49.912704

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "87efceb6745b"
down_revision = "398150912a3f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "entity",
        sa.Column(
            "ready_for",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.create_foreign_key(
        "fk_ready_for", "entity", "task_type", ["ready_for"], ["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("fk_ready_for", "entity", type_="foreignkey")
    op.drop_column("entity", "ready_for")
    # ### end Alembic commands ###



from gevent import monkey

monkey.patch_all()

from flask import Flask, jsonify
from flask_jwt_extended import (
    get_jwt_identity,
    jwt_required,
    verify_jwt_in_request,
    JWTManager,
)
from flask_socketio import SocketIO, disconnect, join_room, emit
from flask_sqlalchemy import SQLAlchemy

from zou.app import config
from zou.app.stores import auth_tokens_store
from zou.app.utils.monitoring import init_monitoring
from zou.app.utils.flask import ORJSONProvider
from zou.app.services import persons_service

server_stats = {"nb_connections": 0}
rooms_data = {}

# Review room helpers


def _get_empty_room(current_frame=0):
    return {
        "people": [],
        "is_playing": False,
        "current_entity_index": None,
        "current_preview_file_id": None,
        "current_frame": current_frame,
        "is_repeating": None,
        "is_laser_mode": None,
        "handle_in": None,
        "handle_out": None,
        "speed": None,
        "comparing": {
            "enable": False,
            "task_type": None,
            "revision": None,
            "mode": "sidebyside",
            "comparison_preview_index": 0,
        },
    }


def _get_room_from_data(data):
    room_id = data["playlist_id"]
    return rooms_data.get(room_id, _get_empty_room()), room_id


def _leave_room(room_id, user_id):
    room = rooms_data.get(room_id, _get_empty_room())
    room["people"] = list(set(room["people"]) - {user_id})
    if len(room["people"]) > 0:
        rooms_data[room_id] = room
    else:
        del rooms_data[room_id]
    emit("preview-room:room-people-updated", room, room=room_id)


def _update_room_playing_status(data, room):
    room["is_playing"] = data.get("is_playing", False)
    room["is_repeating"] = data.get("is_repeating", False)
    room["is_laser_mode"] = data.get("is_laser_mode", False)
    room["current_entity_index"] = data["current_entity_index"]
    room["handle_in"] = data.get("handle_in", None)
    room["handle_out"] = data.get("handle_out", None)
    if "current_preview_file_id" in data:
        room["current_preview_file_id"] = data["current_preview_file_id"]
    if "current_frame" in data:
        room["current_frame"] = data["current_frame"]
    if "comparing" in data:
        room["comparing"] = data["comparing"]
    if "speed" in data:
        room["speed"] = data["speed"]
    return room


# Database helpers


def get_redis_url():
    redis_host = config.KEY_VALUE_STORE["host"]
    redis_port = config.KEY_VALUE_STORE["port"]
    db_index = config.KV_EVENTS_DB_INDEX
    redis_password = config.KEY_VALUE_STORE["password"]
    if redis_password:
        return "redis://:%s@%s:%s/%s" % (
            redis_password,
            redis_host,
            redis_port,
            db_index,
        )
    else:
        return "redis://%s:%s/%s" % (redis_host, redis_port, db_index)


# Routes


def set_info_routes(socketio, app):
    @app.route("/", methods=["GET"])
    def index():
        return jsonify({"name": "%s Event stream" % config.APP_NAME})

    @app.route("/stats", methods=["GET"])
    def stats():
        return jsonify(server_stats)


def set_application_routes(socketio, app):
    @socketio.on("connect", namespace="/events")
    def connected():
        try:
            verify_jwt_in_request()
            server_stats["nb_connections"] += 1
            app.logger.info("New websocket client connected")
        except Exception:
            app.logger.info("New websocket client failed to connect")
            disconnect()
            return False

    @socketio.on("disconnect", namespace="/events")
    def disconnected():
        try:
            verify_jwt_in_request()
            user_id = get_jwt_identity()
            # needed to be able to clear empty rooms
            tmp_rooms_data = dict(rooms_data)
            for room_id in tmp_rooms_data:
                _leave_room(room_id, user_id)
            server_stats["nb_connections"] -= 1
            app.logger.info("Websocket client disconnected")
        except Exception:
            pass

    @socketio.on_error("/events")
    def on_error(error):
        server_stats["nb_connections"] -= 1
        if server_stats["nb_connections"] < 0:
            server_stats["nb_connections"] = 0
        app.logger.error(error)


def set_playlist_room_routes(socketio, app):
    @app.route("/rooms", methods=["GET", "POST"])
    @jwt_required()
    def rooms():
        return jsonify({"name": "%s Review rooms" % config.APP_NAME})

    @socketio.on("preview-room:open-playlist", namespace="/events")
    @jwt_required()
    def on_open_playlist(data):
        """
        when a person opens the playlist page he immediately enters the
        websocket room. This way he can see in live which people are in the
        review room. The user still has to explicitly enter the review room
        to actually be in sync with the other users
        """
        room, room_id = _get_room_from_data(data)
        rooms_data[room_id] = room
        join_room(room_id)
        emit("preview-room:room-people-updated", room, room=room_id)

    @socketio.on("preview-room:join", namespace="/events")
    @jwt_required()
    def on_join(data):
        """
        When a person joins the review room, we notify all its members that a
        new person is added to the room.
        """
        user_id = get_jwt_identity()
        room, room_id = _get_room_from_data(data)
        if len(room["people"]) == 0:
            _update_room_playing_status(data, room)
        room["people"] = list(set(room["people"] + [user_id]))
        rooms_data[room_id] = room
        emit("preview-room:room-people-updated", room, room=room_id)

    @socketio.on("preview-room:leave", namespace="/events")
    @jwt_required()
    def on_leave(data):
        user_id = get_jwt_identity()
        room_id = data["playlist_id"]
        _leave_room(room_id, user_id)

    @socketio.on("preview-room:update-playing-status", namespace="/events")
    @jwt_required()
    def on_playing_status_updated(data, only_newcomer=False):
        room, room_id = _get_room_from_data(data)
        rooms_data[room_id] = _update_room_playing_status(data, room)
        event_data = {"only_newcomer": only_newcomer, **rooms_data[room_id]}
        emit("preview-room:room-updated", event_data, room=room_id)

    @socketio.on("preview-room:add-annotation", namespace="/events")
    @jwt_required()
    def on_add_annotation(data):
        room_id = data["playlist_id"]
        emit("preview-room:add-annotation", data, room=room_id)

    @socketio.on("preview-room:remove-annotation", namespace="/events")
    @jwt_required()
    def on_remove_annotation(data):
        room_id = data["playlist_id"]
        emit("preview-room:remove-annotation", data, room=room_id)

    @socketio.on("preview-room:update-annotation", namespace="/events")
    @jwt_required()
    def on_update_annotation(data):
        room_id = data["playlist_id"]
        emit("preview-room:update-annotation", data, room=room_id)

    @socketio.on("preview-room:change-version", namespace="/events")
    @jwt_required()
    def on_change_version(data):
        room_id = data["playlist_id"]
        emit("preview-room:change-version", data, room=room_id)

    return app


def create_app():
    redis_url = get_redis_url()
    socketio = SocketIO(
        logger=True, cors_allowed_origins=[], cors_credentials=False
    )
    app = Flask(__name__)
    app.json = ORJSONProvider(app)
    app.config.from_object(config)
    init_monitoring(app)
    db = SQLAlchemy(app)
    app.extensions["sqlalchemy"].db = db
    set_info_routes(socketio, app)
    set_application_routes(socketio, app)
    set_playlist_room_routes(socketio, app)
    socketio.init_app(app, message_queue=redis_url, async_mode="gevent")
    return (app, socketio)


def set_auth(app):
    jwt = JWTManager(app)  # JWT auth tokens

    @jwt.token_in_blocklist_loader
    def check_if_token_is_revoked(_, payload):
        identity_type = payload.get("identity_type")
        if identity_type == "person":
            return auth_tokens_store.is_revoked(payload["jti"])
        elif identity_type in ["bot", "person_api"]:
            return persons_service.is_jti_revoked(payload["jti"])
        else:
            return True


(app, socketio) = create_app()
set_auth(app)


if __name__ == "__main__":
    socketio.run(
        app,
        debug=False,
        host=config.EVENT_STREAM_HOST,
        port=config.EVENT_STREAM_PORT,
    )



"""empty message

Revision ID: 40dea9555940
Revises: 2adc020885fa
Create Date: 2019-05-09 18:19:12.766386

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "40dea9555940"
down_revision = "2adc020885fa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("news_preview_file_id_fkey", "news", type_="foreignkey")
    op.create_foreign_key(
        None, "news", "preview_file", ["preview_file_id"], ["id"]
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "news", type_="foreignkey")
    op.create_foreign_key(
        "news_preview_file_id_fkey",
        "news",
        "task",
        ["preview_file_id"],
        ["id"],
    )
    # ### end Alembic commands ###



"""empty message

Revision ID: ee2373fbe3a4
Revises: 99825b9cc778
Create Date: 2018-05-24 11:24:22.019457

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
from sqlalchemy.dialects import postgresql
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "ee2373fbe3a4"
down_revision = "99825b9cc778"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "asset_instance_link",
        sa.Column(
            "entity_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column(
            "asset_instance_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["asset_instance_id"],
            ["asset_instance.id"],
        ),
        sa.ForeignKeyConstraint(
            ["entity_id"],
            ["entity.id"],
        ),
        sa.PrimaryKeyConstraint("entity_id", "asset_instance_id"),
    )
    op.add_column(
        "asset_instance",
        sa.Column(
            "scene_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.alter_column(
        "asset_instance",
        "entity_id",
        existing_type=postgresql.UUID(),
        nullable=True,
    )
    op.alter_column(
        "asset_instance",
        "entity_type_id",
        existing_type=postgresql.UUID(),
        nullable=True,
    )
    op.create_index(
        op.f("ix_asset_instance_scene_id"),
        "asset_instance",
        ["scene_id"],
        unique=False,
    )
    op.drop_constraint(
        "asset_instance_name_uc", "asset_instance", type_="unique"
    )
    op.create_unique_constraint(
        "asset_instance_name_uc", "asset_instance", ["scene_id", "name"]
    )
    op.drop_constraint("asset_instance_uc", "asset_instance", type_="unique")
    op.create_unique_constraint(
        "asset_instance_uc",
        "asset_instance",
        ["asset_id", "scene_id", "number"],
    )
    op.drop_index("ix_asset_instance_entity_id", table_name="asset_instance")
    op.drop_index(
        "ix_asset_instance_entity_type_id", table_name="asset_instance"
    )
    op.create_foreign_key(
        None, "asset_instance", "entity", ["scene_id"], ["id"]
    )
    op.add_column(
        "output_file",
        sa.Column(
            "temporal_entity_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=True,
        ),
    )
    op.drop_constraint("output_file_uc", "output_file", type_="unique")
    op.create_unique_constraint(
        "output_file_uc",
        "output_file",
        [
            "name",
            "entity_id",
            "asset_instance_id",
            "output_type_id",
            "task_type_id",
            "temporal_entity_id",
            "representation",
            "revision",
        ],
    )
    op.create_foreign_key(
        None, "output_file", "entity", ["temporal_entity_id"], ["id"]
    )
    op.drop_column("output_file", "uploaded_movie_name")
    op.drop_column("output_file", "uploaded_movie_url")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "output_file",
        sa.Column(
            "uploaded_movie_url",
            sa.VARCHAR(length=600),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.add_column(
        "output_file",
        sa.Column(
            "uploaded_movie_name",
            sa.VARCHAR(length=150),
            autoincrement=False,
            nullable=True,
        ),
    )
    op.drop_constraint(None, "output_file", type_="foreignkey")
    op.drop_constraint("output_file_uc", "output_file", type_="unique")
    op.create_unique_constraint(
        "output_file_uc",
        "output_file",
        [
            "name",
            "entity_id",
            "output_type_id",
            "task_type_id",
            "representation",
            "revision",
        ],
    )
    op.drop_column("output_file", "temporal_entity_id")
    op.drop_constraint(None, "asset_instance", type_="foreignkey")
    op.create_index(
        "ix_asset_instance_entity_type_id",
        "asset_instance",
        ["entity_type_id"],
        unique=False,
    )
    op.create_index(
        "ix_asset_instance_entity_id",
        "asset_instance",
        ["entity_id"],
        unique=False,
    )
    op.drop_constraint("asset_instance_uc", "asset_instance", type_="unique")
    op.create_unique_constraint(
        "asset_instance_uc",
        "asset_instance",
        ["asset_id", "entity_id", "number"],
    )
    op.drop_constraint(
        "asset_instance_name_uc", "asset_instance", type_="unique"
    )
    op.create_unique_constraint(
        "asset_instance_name_uc", "asset_instance", ["entity_id", "name"]
    )
    op.drop_index(
        op.f("ix_asset_instance_scene_id"), table_name="asset_instance"
    )
    op.alter_column(
        "asset_instance",
        "entity_type_id",
        existing_type=postgresql.UUID(),
        nullable=False,
    )
    op.alter_column(
        "asset_instance",
        "entity_id",
        existing_type=postgresql.UUID(),
        nullable=False,
    )
    op.drop_column("asset_instance", "scene_id")
    op.drop_table("asset_instance_link")
    # ### end Alembic commands ###



"""Add is feedback request column to task types

Revision ID: aa0a60033106
Revises: 7b1f765677d8
Create Date: 2022-01-31 12:39:25.332258

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "aa0a60033106"
down_revision = "7b1f765677d8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_status",
        sa.Column("is_feedback_request", sa.Boolean(), nullable=True),
    )
    op.create_index(
        op.f("ix_task_status_is_feedback_request"),
        "task_status",
        ["is_feedback_request"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_task_status_is_feedback_request"), table_name="task_status"
    )
    op.drop_column("task_status", "is_feedback_request")
    # ### end Alembic commands ###



from zou.app.models.milestone import Milestone
from zou.app.services import user_service

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource


class MilestonesResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, Milestone)

    def check_create_permissions(self, milestone):
        user_service.check_manager_project_access(milestone["project_id"])


class MilestoneResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, Milestone)

    def check_read_permissions(self, milestone):
        user_service.check_project_access(milestone["project_id"])
        user_service.block_access_to_vendor()

    def check_update_permissions(self, milestone, data):
        user_service.check_manager_project_access(milestone["project_id"])



from flask import current_app

from zou.app.models.project import Project
from zou.app.models.entity import Entity
from zou.app.services import shots_service, persons_service
from zou.app.blueprints.source.shotgun.base import (
    BaseImportShotgunResource,
    ImportRemoveShotgunBaseResource,
)
from zou.app.blueprints.source.shotgun.exception import (
    ShotgunEntryImportFailed,
)


class ImportShotgunSequencesResource(BaseImportShotgunResource):
    def prepare_import(self):
        self.sequence_type = shots_service.get_sequence_type()
        self.project_map = Project.get_id_map(field="name")
        self.current_user_id = persons_service.get_current_user()["id"]

    def get_episode(self, sg_sequence):
        sg_episode = sg_sequence.get("episode", {"id": None})
        if sg_episode is not None:
            episode_sg_id = sg_episode.get("id", None)
            return self.get_episode_id(episode_sg_id)
        else:
            return None

    def extract_data(self, sg_sequence):
        project_id = self.get_project(sg_sequence)
        episode_id = self.get_episode(sg_sequence)
        if project_id is None:
            raise ShotgunEntryImportFailed
        return {
            "name": sg_sequence["code"],
            "shotgun_id": sg_sequence["id"],
            "description": sg_sequence["description"],
            "project_id": project_id,
            "parent_id": episode_id,
            "entity_type_id": self.sequence_type["id"],
        }

    def get_project(self, sg_sequence):
        project_id = None
        if sg_sequence["project"] is not None:
            project_name = sg_sequence["project"]["name"]
            project_id = self.project_map.get(project_name, None)
        return project_id

    def import_entry(self, data):
        sequence = Entity.get_by(
            shotgun_id=data["shotgun_id"],
            entity_type_id=self.sequence_type["id"],
        )

        similar_sequence = Entity.get_by(
            name=data["name"],
            parent_id=data["parent_id"],
            project_id=data["project_id"],
            entity_type_id=self.sequence_type["id"],
        )

        if sequence is None and similar_sequence is None:
            sequence = Entity.create(**data, created_by=self.current_user_id)
            current_app.logger.info("Sequence created: %s" % sequence)

        elif sequence is not None:
            if similar_sequence is None:
                sequence.update(data)
                sequence.save()
            else:
                sequence.update(
                    {
                        "description": data["description"],
                        "shotgun_id": data["shotgun_id"],
                    }
                )
                sequence.save()
            current_app.logger.info("Sequence updated: %s" % sequence)

        return sequence


class ImportRemoveShotgunSequenceResource(ImportRemoveShotgunBaseResource):
    def __init__(self):
        ImportRemoveShotgunBaseResource.__init__(
            self,
            Entity,
            entity_type_id=shots_service.get_sequence_type()["id"],
        )



"""Add table Studio

Revision ID: 1bb55759146f
Revises: be56dc0fb760
Create Date: 2024-05-31 15:30:17.189541

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils
import sqlalchemy_utils
import uuid

# revision identifiers, used by Alembic.
revision = "1bb55759146f"
down_revision = "be56dc0fb760"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "studio",
        sa.Column("name", sa.String(length=80), nullable=False),
        sa.Column("color", sa.String(length=7), nullable=False),
        sa.Column("archived", sa.Boolean(), nullable=True),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            default=uuid.uuid4,
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "studio_id",
                sqlalchemy_utils.types.uuid.UUIDType(binary=False),
                default=uuid.uuid4,
                nullable=True,
            )
        )
        batch_op.create_index(
            batch_op.f("ix_person_studio_id"), ["studio_id"], unique=False
        )
        batch_op.create_foreign_key(
            "person_studio_id_fkey", "studio", ["studio_id"], ["id"]
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("person", schema=None) as batch_op:
        batch_op.drop_constraint("person_studio_id_fkey", type_="foreignkey")
        batch_op.drop_index(batch_op.f("ix_person_studio_id"))
        batch_op.drop_column("studio_id")

    op.drop_table("studio")
    # ### end Alembic commands ###



from sqlalchemy_utils import UUIDType, ChoiceType
from sqlalchemy.dialects.postgresql import JSONB

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin

department_metadata_descriptor_link = db.Table(
    "department_metadata_descriptor_link",
    db.Column(
        "metadata_descriptor_id",
        UUIDType(binary=False),
        db.ForeignKey("metadata_descriptor.id"),
    ),
    db.Column(
        "department_id", UUIDType(binary=False), db.ForeignKey("department.id")
    ),
)

METADATA_DESCRIPTOR_TYPES = [
    ("string", "String"),
    ("number", "Number"),
    ("list", "List"),
    ("taglist", "Taglist"),
    ("boolean", "Boolean"),
    ("checklist", "Checklist"),
]


class MetadataDescriptor(db.Model, BaseMixin, SerializerMixin):
    """
    This models allow to identify which metadata are available for a given
    project and a given entity type.
    """

    project_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("project.id"),
        nullable=False,
        index=True,
    )
    entity_type = db.Column(db.String(60), nullable=False, index=True)
    name = db.Column(db.String(120), nullable=False)
    data_type = db.Column(ChoiceType(METADATA_DESCRIPTOR_TYPES))
    field_name = db.Column(db.String(120), nullable=False)
    choices = db.Column(JSONB)
    for_client = db.Column(db.Boolean(), default=False, index=True)
    departments = db.relationship(
        "Department", secondary=department_metadata_descriptor_link
    )

    __table_args__ = (
        db.UniqueConstraint(
            "project_id", "entity_type", "name", name="metadata_descriptor_uc"
        ),
        db.UniqueConstraint(
            "project_id",
            "entity_type",
            "field_name",
            name="metadata_descriptor_uc2",
        ),
    )

    def __repr__(self):
        return "<MetadataDescriptor %s>" % self.id



"""drop task_status.is_reviewable

Revision ID: b8c0a0f9d054
Revises: 6eeaff945706
Create Date: 2022-04-29 15:04:36.498586

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b8c0a0f9d054"
down_revision = "6eeaff945706"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_status", "is_reviewable")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_status",
        sa.Column(
            "is_reviewable", sa.BOOLEAN(), autoincrement=False, nullable=True
        ),
    )
    # ### end Alembic commands ###



"""Add indices

Revision ID: 9010a64e5a2d
Revises: 306266361f4f
Create Date: 2019-11-27 21:31:33.868141

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "9010a64e5a2d"
down_revision = "306266361f4f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(
        op.f("ix_playlist_episode_id"),
        "playlist",
        ["episode_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_playlist_for_client"),
        "playlist",
        ["for_client"],
        unique=False,
    )
    op.create_index(
        op.f("ix_playlist_project_id"),
        "playlist",
        ["project_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_task_assigner_id"), "task", ["assigner_id"], unique=False
    )
    op.create_index(
        op.f("ix_task_task_status_id"),
        "task",
        ["task_status_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_task_task_type_id"), "task", ["task_type_id"], unique=False
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_task_task_type_id"), table_name="task")
    op.drop_index(op.f("ix_task_task_status_id"), table_name="task")
    op.drop_index(op.f("ix_task_assigner_id"), table_name="task")
    op.drop_index(op.f("ix_playlist_project_id"), table_name="playlist")
    op.drop_index(op.f("ix_playlist_for_client"), table_name="playlist")
    op.drop_index(op.f("ix_playlist_episode_id"), table_name="playlist")
    # ### end Alembic commands ###



"""add missing unique constraints

Revision ID: fb87feaaa094
Revises: 1cb44194db49
Create Date: 2021-02-03 19:02:51.347109

"""

from alembic import op


# revision identifiers, used by Alembic.
revision = "fb87feaaa094"
down_revision = "1cb44194db49"
branch_labels = None
depends_on = None


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_unique_constraint(
        "entity_link_uc", "entity_link", ["entity_in_id", "entity_out_id"]
    )
    op.create_unique_constraint(
        "schedule_item_uc",
        "schedule_item",
        ["project_id", "task_type_id", "object_id"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("schedule_item_uc", "schedule_item", type_="unique")
    op.drop_constraint("entity_link_uc", "entity_link", type_="unique")
    # ### end Alembic commands ###



"""Add nb assets ready column

Revision ID: a66508788c53
Revises: 1e150c2cea4d
Create Date: 2021-11-23 00:07:43.717653

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "a66508788c53"
down_revision = "1e150c2cea4d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task", sa.Column("nb_assets_ready", sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task", "nb_assets_ready")
    # ### end Alembic commands ###



"""empty message

Revision ID: a23682ccc1f1
Revises: 9bd17364fc18
Create Date: 2018-04-20 10:39:31.976959

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "a23682ccc1f1"
down_revision = "9bd17364fc18"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "desktop_login_log",
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("updated_at", sa.DateTime(), nullable=True),
        sa.Column(
            "person_id",
            sqlalchemy_utils.types.uuid.UUIDType(binary=False),
            nullable=False,
        ),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["person_id"],
            ["person.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_desktop_login_log_person_id"),
        "desktop_login_log",
        ["person_id"],
        unique=False,
    )
    op.add_column(
        "asset_instance", sa.Column("active", sa.Boolean(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("asset_instance", "active")
    op.drop_index(
        op.f("ix_desktop_login_log_person_id"), table_name="desktop_login_log"
    )
    op.drop_table("desktop_login_log")
    # ### end Alembic commands ###



"""add is_done field to the task model

Revision ID: ca28796a2a62
Revises: 971dbf5a0faf
Create Date: 2024-07-19 15:24:21.064991

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin
from sqlalchemy_utils import UUIDType, ChoiceType


# revision identifiers, used by Alembic.
revision = "ca28796a2a62"
down_revision = "971dbf5a0faf"
branch_labels = None
depends_on = None


def upgrade():
    base = declarative_base()

    TYPES = [
        ("comment", "Comment"),
        ("mention", "Mention"),
        ("assignation", "Assignation"),
        ("reply", "Reply"),
        ("reply-mention", "Reply Mention"),
    ]

    class Notification(base, BaseMixin):
        """
        A notification is stored each time a comment is posted.
        """

        __tablename__ = "notification"
        read = sa.Column(sa.Boolean, nullable=False, default=False)
        change = sa.Column(sa.Boolean, nullable=False, default=False)
        type = sa.Column(ChoiceType(TYPES), nullable=False)
        person_id = sa.Column(
            UUIDType(binary=False),
            nullable=False,
            index=True,
        )
        author_id = sa.Column(
            UUIDType(binary=False),
            nullable=False,
            index=True,
        )
        comment_id = sa.Column(
            UUIDType(binary=False),
            nullable=True,
            index=True,
        )
        task_id = sa.Column(
            UUIDType(binary=False),
            nullable=False,
            index=True,
        )
        reply_id = sa.Column(UUIDType(binary=False), nullable=True, index=True)

        __table_args__ = (
            sa.UniqueConstraint(
                "person_id",
                "author_id",
                "comment_id",
                "reply_id",
                "type",
                name="notification_uc",
            ),
        )

    bind = op.get_bind()
    session = orm.Session(bind=bind)
    session.query(Notification).where(Notification.type == None).update(
        {Notification.type: "comment"}
    )
    session.commit()

    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("notification", schema=None) as batch_op:
        batch_op.alter_column(
            "type", existing_type=sa.VARCHAR(length=255), nullable=False
        )

    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("done_date", sa.DateTime(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("task", schema=None) as batch_op:
        batch_op.drop_column("done_date")

    with op.batch_alter_table("notification", schema=None) as batch_op:
        batch_op.alter_column(
            "type", existing_type=sa.VARCHAR(length=255), nullable=True
        )

    # ### end Alembic commands ###



"""empty message

Revision ID: e1ef93f40d3d
Revises: de8a3de227ef
Create Date: 2018-12-03 12:35:01.505576

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e1ef93f40d3d"
down_revision = "de8a3de227ef"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_status",
        sa.Column("is_artist_allowed", sa.Boolean(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_status", "is_artist_allowed")
    # ### end Alembic commands ###



"""task_status new column is_default

Revision ID: d80267806131
Revises: b8c0a0f9d054
Create Date: 2022-05-03 14:43:47.946725

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base
from zou.migrations.utils.base import BaseMixin

# revision identifiers, used by Alembic.
revision = "d80267806131"
down_revision = "b8c0a0f9d054"
branch_labels = None
depends_on = None


class TaskStatus(declarative_base(), BaseMixin):
    """
    Describe the state of a task. A status marked as reviewable expects a
    preview file linked to relate comment.
    """

    __tablename__ = "task_status"
    name = sa.Column(sa.String(40), nullable=False)
    short_name = sa.Column(
        sa.String(10), unique=True, nullable=False, index=True
    )
    color = sa.Column(sa.String(7), nullable=False)

    is_done = sa.Column(sa.Boolean(), default=False, index=True)
    is_artist_allowed = sa.Column(sa.Boolean(), default=True)
    is_client_allowed = sa.Column(sa.Boolean(), default=True)
    is_retake = sa.Column(sa.Boolean(), default=False)
    is_feedback_request = sa.Column(sa.Boolean(), default=False, index=True)
    is_default = sa.Column(sa.Boolean(), default=False, index=True)
    shotgun_id = sa.Column(sa.Integer)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_status", sa.Column("is_default", sa.Boolean(), nullable=True)
    )
    op.create_index(
        op.f("ix_task_status_is_default"),
        "task_status",
        ["is_default"],
        unique=True,
    )
    # ### end Alembic commands ###

    # Set the todo task status is_default to True
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    todo_task_status = (
        session.query(TaskStatus)
        .filter(TaskStatus.name == "Todo")
        .one_or_none()
    )
    if todo_task_status:
        todo_task_status.is_default = True
        session.commit()


def downgrade():
    # Set the todo task status to default short_name/name/color
    bind = op.get_bind()
    session = orm.Session(bind=bind)
    todo_task_status = (
        session.query(TaskStatus).filter(TaskStatus.is_default == True).one()
    )
    todo_task_status.name = "Todo"
    todo_task_status.short_name = "todo"
    todo_task_status.color = "#FFFFFF"
    session.commit()

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_task_status_is_default"), table_name="task_status")
    op.drop_column("task_status", "is_default")
    # ### end Alembic commands ###



"""Add is client allowed flag to task status

Revision ID: 5b7fa3e51701
Revises: 6d1b2c60f58b
Create Date: 2019-11-04 15:37:12.261423

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5b7fa3e51701"
down_revision = "6d1b2c60f58b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "task_status",
        sa.Column(
            "is_client_allowed", sa.Boolean(), nullable=True, default=False
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_status", "is_client_allowed")
    # ### end Alembic commands ###



from ua_parser import user_agent_parser
from werkzeug.user_agent import UserAgent
from werkzeug.utils import cached_property
from flask.json.provider import JSONProvider
from flask import make_response, request, abort

import orjson

orjson_options = orjson.OPT_NON_STR_KEYS


def output_json(data, code, headers=None):
    """Makes a Flask response with a JSON encoded body"""
    dumped = orjson.dumps(data, option=orjson_options)

    resp = make_response(dumped, code)
    resp.headers.extend(headers or {})
    return resp


class ORJSONProvider(JSONProvider):
    def __init__(self, *args, **kwargs):
        self.options = kwargs
        super().__init__(*args, **kwargs)

    def loads(self, s, **kwargs):
        return orjson.loads(s)

    def dumps(self, obj, **kwargs):
        return orjson.dumps(obj, option=orjson_options).decode("utf-8")


class ParsedUserAgent(UserAgent):
    @cached_property
    def _details(self):
        return user_agent_parser.Parse(self.string)

    @property
    def platform(self):
        return self._details["os"]["family"]

    @property
    def browser(self):
        return self._details["user_agent"]["family"]

    @property
    def version(self):
        return ".".join(
            part
            for key in ("major", "minor", "patch")
            if (part := self._details["user_agent"][key]) is not None
        )


def is_from_browser(user_agent):
    return user_agent.browser in [
        "Brave",
        "Chrome",
        "Chrome Mobile",
        "Chrome Mobile iOS",
        "Edge",
        "Firefox",
        "Mobile Safari",
        "Opera",
        "Safari",
        "Vivaldi",
    ]


def wrong_auth_handler(identity_user=None):
    if request.path not in ["/auth/login", "/auth/logout"]:
        raise abort(401)
    else:
        return identity_user



"""add file size field to preview file

Revision ID: 1cb44194db49
Revises: cf6cec6d6bf5
Create Date: 2021-01-20 10:58:07.202278

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "1cb44194db49"
down_revision = "cf6cec6d6bf5"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "preview_file", sa.Column("file_size", sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("preview_file", "file_size")
    # ### end Alembic commands ###



"""empty message

Revision ID: 2adc020885fa
Revises: 892b264937ec
Create Date: 2019-05-08 16:13:58.958576

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "2adc020885fa"
down_revision = "892b264937ec"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "news",
        "preview_file_id",
        existing_type=postgresql.UUID(),
        nullable=True,
    )
    op.drop_index("ix_news_person_id", table_name="news")
    op.drop_constraint("news_person_id_fkey", "news", type_="foreignkey")
    op.drop_column("news", "person_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "news",
        sa.Column(
            "person_id", postgresql.UUID(), autoincrement=False, nullable=False
        ),
    )
    op.create_foreign_key(
        "news_person_id_fkey", "news", "person", ["person_id"], ["id"]
    )
    op.create_index("ix_news_person_id", "news", ["person_id"], unique=False)
    op.alter_column(
        "news",
        "preview_file_id",
        existing_type=postgresql.UUID(),
        nullable=False,
    )
    # ### end Alembic commands ###



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class ChatParticipant(db.Model):
    __tablename__ = "chat_participant"
    chat_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("chat.id"),
        primary_key=True,
    )
    person_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        primary_key=True,
    )


class Chat(db.Model, BaseMixin, SerializerMixin):
    """
    Message shared in the entity chat feeds.
    """

    object_id = db.Column(UUIDType(binary=False), nullable=False, index=True)
    object_type = db.Column(
        db.String(80), nullable=False, index=True, default="entity"
    )
    last_message = db.Column(db.DateTime, nullable=True)
    participants = db.relationship(
        "Person", secondary="chat_participant", lazy="joined"
    )

    def __repr__(self):
        return "<Message of %s>" % self.object_id

    def present(self):
        return {
            "id": str(self.id),
            "object_id": str(self.object_id),
            "last_message": self.last_message,
        }



from zou.app.models.entity import EntityLink
from zou.app.utils import fields

from zou.app.blueprints.crud.base import BaseModelResource, BaseModelsResource
from zou.app.services.exception import (
    EntityLinkNotFoundException,
    WrongParameterException,
)


class EntityLinksResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, EntityLink)


class EntityLinkResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, EntityLink)

    def get_model_or_404(self, instance_id):
        if not fields.is_valid_id(instance_id):
            raise WrongParameterException("Malformed ID.")
        instance = self.model.get_by(id=instance_id)
        if instance is None:
            raise EntityLinkNotFoundException
        return instance



from zou.app.blueprints.export.csv.base import BaseCsvExport

from zou.app.models.department import Department
from zou.app.models.task_type import TaskType


class TaskTypesCsvExport(BaseCsvExport):
    def __init__(self):
        BaseCsvExport.__init__(self)

        self.name = "task_types_export"

    def build_headers(self):
        return ["Department", "Name"]

    def build_query(self):
        query = TaskType.query.order_by(Department.name, TaskType.name)
        query = query.join(Department, TaskType.department_id == Department.id)
        query = query.add_columns(Department.name)
        return query

    def build_row(self, task_type_row):
        (task_type, department_name) = task_type_row
        return [department_name, task_type.name]



import slugify

from flask import request, send_file as flask_send_file
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app import config
from zou.app.mixin import ArgsMixin

from zou.app.services import (
    entities_service,
    playlists_service,
    persons_service,
    preview_files_service,
    projects_service,
    shots_service,
    user_service,
)
from zou.app.stores import file_store, queue_store
from zou.app.utils import fs, permissions
from zou.utils.movie import EncodingParameters


class ProjectPlaylistsResource(Resource, ArgsMixin):
    """
    Retrieve all playlists related to given project.
    Result is paginated and can be sorted.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Retrieve all playlists related to given project.
        ---
        tags:
        - Playlists
        description: Result is paginated and can be sorted.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All playlists related to given project
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        page = self.get_page()
        sort_by = self.get_sort_by()
        task_type_id = self.get_text_parameter("task_type_id")
        return playlists_service.all_playlists_for_project(
            project_id,
            for_client=permissions.has_client_permissions(),
            page=page,
            sort_by=sort_by,
            task_type_id=task_type_id,
        )


class EpisodePlaylistsResource(Resource, ArgsMixin):
    """
    Retrieve all playlists related to given episode.
    The full list is returned because the number of playlists in an episode is not that big.
    """

    @jwt_required()
    def get(self, project_id, episode_id):
        """
        Retrieve all playlists related to given episode.
        ---
        tags:
        - Playlists
        description: The full list is returned because the number of playlists in an episode is not that big.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: episode_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All playlists related to given episode
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        page = self.get_page()
        sort_by = self.get_sort_by()
        task_type_id = self.get_text_parameter("task_type_id")
        if episode_id not in ["main", "all"]:
            shots_service.get_episode(episode_id)
        return playlists_service.all_playlists_for_episode(
            project_id,
            episode_id,
            for_client=permissions.has_client_permissions(),
            page=page,
            sort_by=sort_by,
            task_type_id=task_type_id,
        )


class ProjectPlaylistResource(Resource):
    """
    Retrieve all playlists related to given project.
    """

    @jwt_required()
    def get(self, project_id, playlist_id):
        """
        Retrieve all playlists related to given project.
        ---
        tags:
        - Playlists
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All playlists related to given project
        """
        user_service.block_access_to_vendor()
        user_service.check_project_access(project_id)
        return playlists_service.get_playlist_with_preview_file_revisions(
            playlist_id
        )


class EntityPreviewsResource(Resource):
    """
    Retrieve all previews related to a given entity.
    It sends them as a dict.
    Keys are related task type ids and values are arrays of preview for this task type.
    """

    @jwt_required()
    def get(self, entity_id):
        """
        Retrieve all previews related to a given entity.
        ---
        tags:
        - Playlists
        description: It sends them as a dict.
                     Keys are related task type ids and values are arrays of preview for this task type.
        parameters:
          - in: path
            name: entity_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description:  All previews related to given entity
        """
        entity = entities_service.get_entity(entity_id)
        user_service.check_project_access(entity["project_id"])
        return playlists_service.get_preview_files_for_entity(entity_id)


class PlaylistDownloadResource(Resource):
    """
    Download given playlist as a .mp4 after given build job is finished.
    """

    @jwt_required()
    def get(self, playlist_id, build_job_id):
        """
        Download given playlist build as .mp4.
        ---
        tags:
        - Playlists
        produces:
          - multipart/form-data
        parameters:
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: build_job_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Given playlist build downloaded as .mp4
            400:
                description: Build not finished, need to retry later
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist, supervisor_access=True)
        project = projects_service.get_project(playlist["project_id"])
        build_job = playlists_service.get_build_job(build_job_id)

        if build_job["status"] != "succeeded":
            return {"error": True, "message": "Build is not finished"}, 400
        else:
            movie_file_path = fs.get_file_path_and_file(
                config,
                file_store.get_local_movie_path,
                file_store.open_movie,
                "playlists",
                build_job_id,
                "mp4",
            )
            context_name = slugify.slugify(project["name"], separator="_")
            if project["production_type"] == "tvshow":
                episode_id = playlist["episode_id"]
                if episode_id is not None:
                    episode = shots_service.get_episode(playlist["episode_id"])
                    episode_name = episode["name"]
                elif playlist["is_for_all"]:
                    episode_name = "all assets"
                else:
                    episode_name = "main pack"
                context_name += "_%s" % slugify.slugify(
                    episode_name, separator="_"
                )
            download_name = "%s_%s_%s.mp4" % (
                slugify.slugify(build_job["created_at"], separator="").replace(
                    "t", "_"
                ),
                context_name,
                slugify.slugify(playlist["name"], separator="_"),
            )
            return flask_send_file(
                movie_file_path,
                conditional=True,
                mimetype="video/mp4",
                as_attachment=True,
                download_name=download_name,
            )


class BuildPlaylistMovieResource(Resource, ArgsMixin):
    """
    Build given playlist as mp4 movie.
    """

    @jwt_required()
    def get(self, playlist_id):
        """
        Build given playlist as mp4 movie.
        ---
        tags:
        - Playlists
        produces:
          - multipart/form-data
        parameters:
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Given playlist built as mp4 movie
        """
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_manager_project_access(playlist["project_id"])

        project = projects_service.get_project(playlist["project_id"])
        (width, height) = preview_files_service.get_preview_file_dimensions(
            project
        )
        fps = preview_files_service.get_preview_file_fps(project)
        full = self.get_bool_parameter("full")
        params = EncodingParameters(width=width, height=height, fps=fps)

        shots = [
            {"preview_file_id": x.get("preview_file_id")}
            for x in playlist["shots"]
        ]

        job = playlists_service.start_build_job(playlist)
        if config.ENABLE_JOB_QUEUE:
            remote = config.ENABLE_JOB_QUEUE_REMOTE
            # remote worker can not access files local to the web app
            assert not remote or config.FS_BACKEND in ["s3", "swift"]

            current_user = persons_service.get_current_user()
            queue_store.job_queue.enqueue(
                playlists_service.build_playlist_job,
                args=(
                    playlist,
                    job,
                    shots,
                    params,
                    current_user["email"],
                    full,
                    remote,
                ),
                job_timeout=int(config.JOB_QUEUE_TIMEOUT),
            )
            return job
        else:
            job = playlists_service.build_playlist_movie_file(
                playlist, job, shots, params, full, remote=False
            )
            return job


class PlaylistZipDownloadResource(Resource):
    """
    Download given playlist as zip.
    """

    @jwt_required()
    def get(self, playlist_id):
        """
        Download given playlist as zip.
        ---
        tags:
        - Playlists
        produces:
          - multipart/form-data
        parameters:
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Given playlist downloaded as zip
                schema:
                    type: file
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist, supervisor_access=True)
        project = projects_service.get_project(playlist["project_id"])
        zip_file_path = playlists_service.build_playlist_zip_file(playlist)

        context_name = slugify.slugify(project["name"], separator="_")
        if project["production_type"] == "tvshow":
            episode_id = playlist["episode_id"]
            if episode_id is not None:
                episode = shots_service.get_episode(playlist["episode_id"])
                episode_name = episode["name"]
            elif playlist["is_for_all"]:
                episode_name = "all assets"
            else:
                episode_name = "main pack"
            context_name += "_%s" % slugify.slugify(
                episode_name, separator="_"
            )
        download_name = "%s_%s.zip" % (
            context_name,
            slugify.slugify(playlist["name"], separator="_"),
        )

        return flask_send_file(
            zip_file_path,
            conditional=True,
            mimetype="application/zip",
            as_attachment=True,
            download_name=download_name,
        )


class BuildJobResource(Resource):
    """
    Retrieve or remove a given build job related to a given playlist.
    """

    @jwt_required()
    def get(self, playlist_id, build_job_id):
        """
        Retrieve build job related to given playlist.
        ---
        tags:
        - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: build_job_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Build job related to given playlist
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist)
        return playlists_service.get_build_job(build_job_id)

    @jwt_required()
    def delete(self, playlist_id, build_job_id):
        """
        Remove given build job related to given playlist.
        ---
        tags:
        - Playlists
        parameters:
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: path
            name: build_job_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
                description: Given build job removed
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist)
        playlists_service.remove_build_job(playlist, build_job_id)
        return "", 204


class ProjectBuildJobsResource(Resource):
    """
    Retrieve all build jobs related to given project.
    It's mainly used for synchronisation purpose.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Retrieve all build jobs related to given project.
        ---
        tags:
        - Playlists
        description: It's mainly used for synchronisation purpose.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All build jobs related to given project
        """
        permissions.check_admin_permissions()
        projects_service.get_project(project_id)
        return playlists_service.get_build_jobs_for_project(project_id)


class ProjectAllPlaylistsResource(Resource, ArgsMixin):
    """
    Retrieve all playlists related to given project.
    It's mainly used for synchronisation purpose.
    """

    @jwt_required()
    def get(self, project_id):
        """
        Retrieve all playlists related to given project.
        ---
        tags:
        - Playlists
        description: It's mainly used for synchronisation purpose.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All playlists related to given project
        """
        permissions.check_admin_permissions()
        projects_service.get_project(project_id)
        page = self.get_page()
        return playlists_service.get_playlists_for_project(project_id, page)


class TempPlaylistResource(Resource, ArgsMixin):
    """
    Retrieve all playlists related to given project.
    It's mainly used for synchronisation purpose.
    """

    @jwt_required()
    def post(self, project_id):
        """
        Retrieve all playlists related to given project.
        ---
        tags:
        - Playlists
        description: It's mainly used for synchronisation purpose.
        parameters:
          - in: path
            name: project_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: All playlists related to given project
        """
        user_service.check_project_access(project_id)
        task_ids = request.json.get("task_ids", [])
        sort = self.get_bool_parameter("sort")
        return (
            playlists_service.generate_temp_playlist(task_ids, sort=sort) or []
        )



from sqlalchemy_utils import UUIDType
from sqlalchemy.dialects.postgresql import JSONB

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


preview_link_table = db.Table(
    "comment_preview_link",
    db.Column(
        "comment",
        UUIDType(binary=False),
        db.ForeignKey("comment.id"),
        primary_key=True,
    ),
    db.Column(
        "preview_file",
        UUIDType(binary=False),
        db.ForeignKey("preview_file.id"),
        primary_key=True,
    ),
)


mentions_table = db.Table(
    "comment_mentions",
    db.Column(
        "comment",
        UUIDType(binary=False),
        db.ForeignKey("comment.id"),
        primary_key=True,
    ),
    db.Column(
        "person",
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        primary_key=True,
    ),
)


department_mentions_table = db.Table(
    "comment_department_mentions",
    db.Column(
        "comment",
        UUIDType(binary=False),
        db.ForeignKey("comment.id"),
        primary_key=True,
    ),
    db.Column(
        "department",
        UUIDType(binary=False),
        db.ForeignKey("department.id"),
        primary_key=True,
    ),
)


acknowledgements_table = db.Table(
    "comment_acknoledgments",
    db.Column(
        "comment",
        UUIDType(binary=False),
        db.ForeignKey("comment.id"),
        primary_key=True,
    ),
    db.Column(
        "person",
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        primary_key=True,
    ),
)


class Comment(db.Model, BaseMixin, SerializerMixin):
    """
    Comment can occur on any object but they are mainly used on tasks.
    In the case of comment tasks, they are linked to a task status and
    eventually to some preview files.
    The status means that comment leads to task status change. The preview file
    means that the comment relates to this preview in the context of the task.
    """

    shotgun_id = db.Column(db.Integer)

    object_id = db.Column(UUIDType(binary=False), nullable=False, index=True)
    object_type = db.Column(db.String(80), nullable=False, index=True)
    text = db.Column(db.Text())
    data = db.Column(JSONB)
    replies = db.Column(JSONB, default=[])
    checklist = db.Column(JSONB)
    pinned = db.Column(db.Boolean)
    links = db.Column(db.ARRAY(db.String()))

    task_status_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("task_status.id"), index=True
    )
    person_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        nullable=False,
        index=True,
    )
    preview_file_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("preview_file.id")
    )
    previews = db.relationship(
        "PreviewFile", secondary=preview_link_table, backref="comments"
    )
    mentions = db.relationship("Person", secondary=mentions_table)
    department_mentions = db.relationship(
        "Department", secondary=department_mentions_table
    )
    acknowledgements = db.relationship(
        "Person", secondary=acknowledgements_table
    )
    attachment_files = db.relationship("AttachmentFile", backref="comment")

    def __repr__(self):
        return "<Comment of %s>" % self.object_id

    def set_preview_files(self, preview_file_ids):
        from zou.app.models.preview_file import PreviewFile

        self.set_many_to_one("previews", PreviewFile, preview_file_ids)

    def set_mentions(self, person_ids):
        from zou.app.models.person import Person

        self.mentions = []
        self.set_many_to_one("mentions", Person, person_ids)

    def set_department_mentions(self, department_ids):
        from zou.app.models.department import Department

        self.mentions = []
        self.set_many_to_one("department_mentions", Department, department_ids)

    def set_acknowledgements(self, person_ids):
        from zou.app.models.person import Person

        self.set_many_to_one("acknowledgements", Person, person_ids)

    def set_attachment_files(self, attachment_file_ids):
        from zou.app.models.attachment_file import AttachmentFile

        self.set_many_to_one(
            "attachment_files", AttachmentFile, attachment_file_ids
        )

    def set_many_to_one(self, field_name, model, model_ids):
        setattr(self, field_name, [])
        for model_id in model_ids:
            instance = model.get(model_id)
            if instance is not None:
                getattr(self, field_name).append(instance)
        self.save()

    @classmethod
    def create_from_import(cls, data):
        is_update = False
        previous_comment = cls.get(data["id"])
        data.pop("type", None)
        preview_file_ids = data.pop("previews", None)
        mention_ids = data.pop("mentions", None)
        department_mention_ids = data.pop("department_mentions", None)
        acknowledgement_ids = data.pop("acknowledgements", None)
        attachment_file_ids = data.pop("attachment_files", None)

        if previous_comment is None:
            previous_comment = cls.create(**data)
            previous_comment.save()
        else:
            is_update = True
            previous_comment.update(data)
            previous_comment.save()

        if preview_file_ids is not None:
            previous_comment.set_preview_files(preview_file_ids)

        if mention_ids is not None:
            previous_comment.set_mentions(mention_ids)

        if department_mention_ids is not None:
            previous_comment.set_department_mentions(department_mention_ids)

        if acknowledgement_ids is not None:
            previous_comment.set_acknowledgements(acknowledgement_ids)

        if attachment_file_ids is not None:
            previous_comment.set_attachment_files(attachment_file_ids)

        return (previous_comment, is_update)



from flask import request, current_app
from flask_restful import Resource
from flask_jwt_extended import jwt_required

from zou.app.utils import fields, permissions
from zou.app.blueprints.source.shotgun.exception import (
    ShotgunEntryImportFailed,
)

from zou.app.services import assets_service, shots_service, tasks_service

from zou.app.services.exception import (
    AssetNotFoundException,
    EpisodeNotFoundException,
    SceneNotFoundException,
    SequenceNotFoundException,
    ShotNotFoundException,
    TaskNotFoundException,
)

from sqlalchemy.exc import IntegrityError, DataError


class BaseImportShotgunResource(Resource):
    def __init__(self):
        Resource.__init__(self)

    @jwt_required()
    def post(self):
        """
        Import shotgun resource.
        ---
        tags:
          - Import
        parameters:
          - in: body
            name: sg_entries
            required: True
            schema:
                type: array
                items:
                    type: object
                    properties:
                        id:
                            type: string
        responses:
            200:
                description: Resource imported
        """
        results = []
        self.sg_entries = request.json

        self.check_permissions()
        self.prepare_import()

        for sg_entry in self.filtered_entries():
            try:
                data = self.extract_data(sg_entry)
                result_entry = self.import_entry(data)
                results.append(result_entry)
            except ShotgunEntryImportFailed as exception:
                current_app.logger.warning(exception)
            except KeyError as exception:
                current_app.logger.warning(exception)
                current_app.logger.error(
                    "Your data is not properly formatted: %s" % sg_entry
                )
            except IntegrityError as exception:
                current_app.logger.error(exception)
                current_app.logger.error(
                    "Data information are duplicated or wrong: %s" % sg_entry
                )
                raise
            except DataError as exception:
                current_app.logger.error(exception)
                current_app.logger.error(
                    "Data cannot be stored (schema error)" % sg_entry
                )
                raise

        self.post_processing()

        return fields.serialize_models(results), 200

    def filtered_entries(self):
        return self.sg_entries

    def check_permissions(self):
        return permissions.check_admin_permissions()

    def prepare_import(self):
        pass

    def extract_data(self, sg_entry):
        pass

    def import_entry(self, data):
        pass

    def post_processing(self):
        pass

    def get_instance_id(self, get_by_sg_id_func, sg_id, exception):
        try:
            return get_by_sg_id_func(sg_id)["id"]
        except exception:
            return None

    def get_task_id(self, task_sg_id):
        return self.get_instance_id(
            tasks_service.get_task_by_shotgun_id,
            task_sg_id,
            TaskNotFoundException,
        )

    def get_asset_id(self, asset_sg_id):
        return self.get_instance_id(
            assets_service.get_asset_by_shotgun_id,
            asset_sg_id,
            AssetNotFoundException,
        )

    def get_shot_id(self, shot_sg_id):
        return self.get_instance_id(
            shots_service.get_shot_by_shotgun_id,
            shot_sg_id,
            ShotNotFoundException,
        )

    def get_scene_id(self, scene_sg_id):
        return self.get_instance_id(
            shots_service.get_scene_by_shotgun_id,
            scene_sg_id,
            SceneNotFoundException,
        )

    def get_sequence_id(self, sequence_sg_id):
        return self.get_instance_id(
            shots_service.get_sequence_by_shotgun_id,
            sequence_sg_id,
            SequenceNotFoundException,
        )

    def get_episode_id(self, episode_sg_id):
        return self.get_instance_id(
            shots_service.get_episode_by_shotgun_id,
            episode_sg_id,
            EpisodeNotFoundException,
        )

    def extract_custom_data(self, sg_shot):
        return {
            k: v
            for k, v in sg_shot.items()
            if self.is_custom_field(k) and v is not None
        }

    def is_custom_field(self, name):
        return name[:3] == "sg_"


class ImportRemoveShotgunBaseResource(Resource):
    def __init__(self, model, delete_func=None, entity_type_id=None):
        Resource.__init__(self)
        self.model = model
        self.delete_func = delete_func
        self.entity_type_id = entity_type_id

    @jwt_required()
    def post(self):
        """
        Import remove instance.
        ---
        tags:
          - Import
        responses:
            204:
                description: Instance removed
        """
        sg_model = request.json
        instance = self.get_instance(sg_model)

        if instance is not None:
            result = {
                "removed_instance_id": str(instance.id),
                "success": self.delete_instance(instance),
            }
        else:
            result = {"success": True}

        return result

    def get_instance(self, sg_model):
        if self.entity_type_id is not None:
            instance = self.model.get_by(
                shotgun_id=sg_model["id"], entity_type_id=self.entity_type_id
            )
        else:
            instance = self.model.get_by(shotgun_id=sg_model["id"])
        return instance

    def delete_instance(self, instance):
        is_success = True
        try:
            if self.delete_func is not None:
                self.delete_func(instance)
            else:
                instance.delete()
        except IntegrityError as exception:
            sg_id = instance.shotgun_id
            current_app.logger.error(str(exception))
            current_app.logger.error(
                "An error occured while deleting model %s." % sg_id
            )
            is_success = False
        return is_success



import traceback
import uuid

from flask import Flask, jsonify, current_app
from flasgger import Swagger
from flask_jwt_extended import JWTManager
from flask_principal import (
    Principal,
    Identity,
    identity_changed,
    RoleNeed,
    UserNeed,
    identity_loaded,
)
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from flask_mail import Mail

from jwt import ExpiredSignatureError
from babel.core import UnknownLocaleError
from meilisearch.errors import (
    MeilisearchApiError,
    MeilisearchCommunicationError,
)

from zou.app import config, swagger
from zou.app.stores import auth_tokens_store, file_store
from zou.app.services.exception import (
    ModelWithRelationsDeletionException,
    PersonNotFoundException,
    WrongIdFormatException,
    WrongParameterException,
    WrongTaskTypeForEntityException,
    UnactiveUserException,
)

from zou.app.utils import cache, fs, logs, monitoring
from zou.app.utils.flask import (
    ParsedUserAgent,
    ORJSONProvider,
    wrong_auth_handler,
)

from zou.app.utils.saml import saml_client_for
from zou.app.utils.fido import get_fido_server

app = Flask(__name__)
app.json = ORJSONProvider(app)
app.request_class.user_agent_class = ParsedUserAgent
app.config.from_object(config)

monitoring.init_monitoring(app)

logs.configure_logs_ovh(app)

db = SQLAlchemy(app)
migrate = Migrate(app, db)  # DB schema migration features
app.extensions["sqlalchemy"].db = db

app.secret_key = app.config["SECRET_KEY"]
jwt = JWTManager(app)  # JWT auth tokens
Principal(app)  # Permissions
cache.cache.init_app(app)  # Function caching
mail = Mail()
mail.init_app(app)  # To send emails
swagger = Swagger(
    app, template=swagger.swagger_template, config=swagger.swagger_config
)

if config.SAML_ENABLED:
    app.extensions["saml_client"] = saml_client_for(config.SAML_METADATA_URL)

app.extensions["fido_server"] = get_fido_server()


@app.teardown_appcontext
def shutdown_session(exception=None):
    db.session.remove()


@app.errorhandler(404)
def page_not_found(error):
    return jsonify(error=True, message=str(error)), 404


@app.errorhandler(WrongIdFormatException)
def id_parameter_format_error(error):
    return (
        jsonify(
            error=True,
            message="One of the ID sent in parameter is not properly formatted.",
        ),
        400,
    )


@app.errorhandler(WrongParameterException)
def wrong_parameter(error):
    return jsonify(error=True, message=str(error), data=error.dict), 400


@app.errorhandler(ExpiredSignatureError)
def wrong_token_signature(error):
    return jsonify(error=True, message=str(error)), 401


@app.errorhandler(ModelWithRelationsDeletionException)
def try_delete_model_with_relations(error):
    return jsonify(error=True, message=str(error)), 400


@app.errorhandler(WrongTaskTypeForEntityException)
def wrong_task_type_for_entity(error):
    return jsonify(error=True, message=str(error)), 400


@app.errorhandler(UnknownLocaleError)
def wrong_locale_label(error):
    return jsonify(error=True, message=str(error)), 400


@app.errorhandler(MeilisearchCommunicationError)
def indexer_not_reachable(error):
    current_app.logger.error("Indexer not reachable")
    return jsonify(error=True, message="Indexer not reachable"), 500


@app.errorhandler(MeilisearchApiError)
def indexer_key_error(error):
    if error.code == "invalid_api_key":
        current_app.logger.error("The indexer key is rejected")
        return jsonify(error=True, message="The indexer key is rejected"), 500
    else:
        raise error


if config.DEBUG:

    @app.errorhandler(Exception)
    def server_error(error):
        stacktrace = traceback.format_exc()
        current_app.logger.error(stacktrace)
        return (
            jsonify(error=True, message=str(error), stacktrace=stacktrace),
            500,
        )


def configure_auth():
    from zou.app.services import persons_service
    from zou.app.services.auth_service import logout

    def check_active_identity(identity, identity_type, jti):
        if not identity.active:
            if identity_type == "person":
                logout(jti=jti)
            current_app.logger.error(
                f"Identity {identity.id} is not active anymore"
            )
            raise UnactiveUserException

    @jwt.token_in_blocklist_loader
    def check_if_token_is_revoked(_, payload):
        identity_type = payload.get("identity_type")
        if identity_type == "person":
            return auth_tokens_store.is_revoked(payload["jti"])
        elif identity_type in ["bot", "person_api"]:
            return persons_service.is_jti_revoked(payload["jti"])
        else:
            return True

    @jwt.user_lookup_loader
    def user_lookup_callback(_, payload):
        identity_type = payload.get("identity_type")
        try:
            identity = persons_service.get_person_raw(payload["sub"])
        except PersonNotFoundException:
            return wrong_auth_handler()
        check_active_identity(identity, identity_type, jti=payload["jti"])
        identity_changed.send(
            current_app._get_current_object(),
            identity=Identity(identity.id, identity_type),
        )
        return identity

    @identity_loaded.connect_via(app)
    def on_identity_loaded(_, identity):
        try:
            if isinstance(identity.id, (str, uuid.UUID)):
                identity.user = persons_service.get_person_raw(identity.id)

                if hasattr(identity.user, "id"):
                    identity.provides.add(UserNeed(identity.user.id))

                if identity.user.role == "admin":
                    identity.provides.add(RoleNeed("admin"))
                    identity.provides.add(RoleNeed("manager"))

                if identity.user.role == "manager":
                    identity.provides.add(RoleNeed("manager"))

                if identity.user.role == "supervisor":
                    identity.provides.add(RoleNeed("supervisor"))

                if identity.user.role == "client":
                    identity.provides.add(RoleNeed("client"))

                if identity.user.role == "vendor":
                    identity.provides.add(RoleNeed("vendor"))

                identity.provides.add(RoleNeed(identity.auth_type))

            return identity

        except Exception as e:
            if isinstance(e, TimeoutError):
                current_app.logger.error("Identity loading timed out")
            if isinstance(e, (PersonNotFoundException, UnactiveUserException)):
                pass
            else:
                current_app.logger.error(e, exc_info=1)
                if hasattr(e, "message"):
                    current_app.logger.error(e.message)
            return wrong_auth_handler()


def load_api(app):
    from zou.app import api

    api.configure(app)

    fs.mkdir_p(app.config["TMP_DIR"])
    configure_auth()


file_store.configure_storages(app)
load_api(app)



import sqlalchemy.orm as orm
from sqlalchemy.inspection import inspect
from zou.app.utils.fields import serialize_value


class SerializerMixin(object):
    """
    Helpers to facilitate JSON serialization of models.
    """

    def is_join(self, attr):
        return hasattr(getattr(self.__class__, attr), "impl") and isinstance(
            getattr(self.__class__, attr).impl,
            orm.attributes.CollectionAttributeImpl,
        )

    def serialize(
        self,
        obj_type=None,
        relations=False,
        milliseconds=False,
        ignored_attrs=[],
    ):
        attrs = inspect(self.__class__).all_orm_descriptors.keys()
        obj_dict = {
            attr: serialize_value(
                getattr(self, attr), milliseconds=milliseconds
            )
            for attr in attrs
            if attr not in ignored_attrs
            and (relations or not self.is_join(attr))
        }
        obj_dict["type"] = obj_type or type(self).__name__
        return obj_dict

    @staticmethod
    def serialize_list(
        models, obj_type=None, relations=False, milliseconds=False, **kwargs
    ):
        return [
            model.serialize(
                obj_type=obj_type,
                relations=relations,
                milliseconds=milliseconds,
                **kwargs
            )
            for model in models
        ]



from flask_restful import Resource
from flask_jwt_extended import jwt_required
from slugify import slugify

from zou.app.utils import date_helpers

from zou.app.services import (
    entities_service,
    files_service,
    names_service,
    playlists_service,
    persons_service,
    projects_service,
    shots_service,
    user_service,
    tasks_service,
)
from zou.app.utils import csv_utils


class PlaylistCsvExport(Resource):
    @jwt_required()
    def get(self, playlist_id):
        """
        Export a given playlist as csv.
        ---
        tags:
          - Export
        parameters:
          - in: path
            name: playlist_id
            required: True
            type: string
            format: UUID
            x-example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            200:
                description: Playlist exported as csv
        """
        user_service.block_access_to_vendor()
        playlist = playlists_service.get_playlist(playlist_id)
        user_service.check_playlist_access(playlist, supervisor_access=True)
        project = projects_service.get_project(playlist["project_id"])
        self.task_type_map = tasks_service.get_task_type_map()
        self.task_status_map = tasks_service.get_task_status_map()
        task_ids = []
        for shot in playlist["shots"]:
            preview_file = files_service.get_preview_file(
                shot["preview_file_id"]
            )
            task_ids.append(preview_file["task_id"])
        self.task_comment_map = tasks_service.get_last_comment_map(task_ids)
        episode = self.get_episode(playlist)

        csv_content = []
        headers = self.build_headers(playlist, project, episode)
        csv_content += headers
        for shot in playlist["shots"]:
            csv_content.append(self.build_row(shot))

        file_name = "%s playlist %s" % (project["name"], playlist["name"])
        return csv_utils.build_csv_response(csv_content, slugify(file_name))

    def build_headers(self, playlist, project, episode=None):
        entity_type = "for assets"
        if playlist["for_entity"] == "shot":
            entity_type = "for shots"
        context_name = project["name"]
        if episode:
            context_name += " - %s" % episode["name"]
        context_name += " | %s" % entity_type
        timezone = persons_service.get_current_user()["timezone"]
        created_at = date_helpers.get_date_string_with_timezone(
            playlist["created_at"], timezone
        )
        headers = [
            [
                "Playlist",
                context_name,
                playlist["name"],
                created_at,
                "",
                "",
                "",
            ],
            ["", "", "", "", "", "", ""],
            [
                "Entity name",
                "Nb Frames",
                "Task Type",
                "Retake count",
                "Revision",
                "Task Status",
                "Last comment author",
                "Last comment date",
                "Last comment",
            ],
        ]
        return headers

    def build_row(self, shot):
        entity = entities_service.get_entity(shot["entity_id"])
        name, _, _ = names_service.get_full_entity_name(shot["entity_id"])
        preview_file = files_service.get_preview_file(shot["preview_file_id"])
        task = tasks_service.get_task(preview_file["task_id"])
        task_type = self.task_type_map[task["task_type_id"]]
        task_status = self.task_status_map[task["task_status_id"]]
        comment = self.task_comment_map.get(task["id"], {})
        author = self.get_author(comment)
        date = self.get_date(comment)
        return [
            name,
            entity.get("nb_frames", ""),
            task_type["name"],
            task["retake_count"],
            preview_file["revision"],
            task_status["name"],
            author,
            date,
            comment.get("text", ""),
        ]

    def get_episode(self, playlist):
        episode = None
        if playlist["episode_id"] is not None:
            episode = shots_service.get_episode(playlist["episode_id"])
        return episode

    def get_author(self, comment):
        author = ""
        person_id = comment.get("person_id", None)
        if person_id is not None:
            person = persons_service.get_person(person_id)
            author = person["full_name"]
        return author

    def get_date(self, comment):
        comment_date = comment.get("date", None)
        if comment_date is not None:
            timezone = persons_service.get_current_user()["timezone"]
            return date_helpers.get_date_string_with_timezone(
                comment_date, timezone
            )
        else:
            return ""



from zou.app.models.custom_action import CustomAction
from zou.app.utils import cache, fields


def clear_custom_action_cache():
    cache.cache.delete_memoized(get_custom_actions)


@cache.memoize_function(120)
def get_custom_actions():
    return fields.serialize_models(CustomAction.get_all())



from sqlalchemy_utils import UUIDType
from sqlalchemy.dialects.postgresql import JSONB

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class AssetInstance(db.Model, BaseMixin, SerializerMixin):
    """
    An asset instance is the representation of an asset in a given shot or
    layout scene. It is useful for complex scenes where an asset needs extra
    treatments only related to the given shot or layout scene.
    An asset can have multiple instances in a scene or in a shot (ex: a sword in
    a battle field).
    """

    asset_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("entity.id"),
        nullable=False,
        index=True,
    )
    name = db.Column(db.String(80))
    number = db.Column(db.Integer())
    description = db.Column(db.String(200))
    active = db.Column(db.Boolean(), default=True)
    data = db.Column(JSONB)

    scene_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("entity.id"), index=True
    )
    target_asset_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("entity.id"), index=True
    )

    __table_args__ = (
        db.UniqueConstraint(
            "asset_id",
            "target_asset_id",
            "scene_id",
            "number",
            name="asset_instance_uc",
        ),
    )

    # Do not use these column. They are deprecated and will be dropped in
    # upcoming version
    entity_id = db.Column(UUIDType(binary=False), db.ForeignKey("entity.id"))
    entity_type_id = db.Column(
        UUIDType(binary=False), db.ForeignKey("entity_type.id")
    )

    def __repr__(self):
        return "<AssetInstance %s>" % self.id



from sqlalchemy_utils import UUIDType

from zou.app import db
from zou.app.models.serializer import SerializerMixin
from zou.app.models.base import BaseMixin


class DesktopLoginLog(db.Model, BaseMixin, SerializerMixin):
    """
    Table to log all desktop session logins. The aim is to build report that
    helps validating presence form.
    """

    person_id = db.Column(
        UUIDType(binary=False),
        db.ForeignKey("person.id"),
        nullable=False,
        index=True,
    )
    date = db.Column(db.DateTime, nullable=False)



"""Add man days fields

Revision ID: 9f8445f9b42c
Revises: e29638428dfd
Create Date: 2019-07-10 19:54:05.160074

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "9f8445f9b42c"
down_revision = "e29638428dfd"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project", sa.Column("man_days", sa.Integer(), nullable=True)
    )
    op.add_column(
        "schedule_item", sa.Column("man_days", sa.Integer(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("schedule_item", "man_days")
    op.drop_column("project", "man_days")
    # ### end Alembic commands ###



"""add is_shared flag to filters

Revision ID: 32f134ff1201
Revises: 57222395f2be
Create Date: 2024-05-17 00:50:00.493167

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "32f134ff1201"
down_revision = "57222395f2be"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("search_filter", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("is_shared", sa.Boolean(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("search_filter", schema=None) as batch_op:
        batch_op.drop_column("is_shared")



"""empty message

Revision ID: f995b28fb749
Revises: 38baa9a23b3d
Create Date: 2018-09-13 13:40:05.558866

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "f995b28fb749"
down_revision = "38baa9a23b3d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project", sa.Column("fps", sa.String(length=10), nullable=True)
    )
    op.add_column(
        "project", sa.Column("ratio", sa.String(length=10), nullable=True)
    )
    op.add_column(
        "project", sa.Column("resolution", sa.String(length=12), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "resolution")
    op.drop_column("project", "ratio")
    op.drop_column("project", "fps")
    # ### end Alembic commands ###



from flask import Blueprint
from zou.app.utils.api import configure_api_from_blueprint

from zou.app.blueprints.projects.resources import (
    AllProjectsResource,
    OpenProjectsResource,
    ProductionTeamResource,
    ProductionTeamRemoveResource,
    ProductionAssetTypeResource,
    ProductionAssetTypeRemoveResource,
    ProductionPreviewBackgroundFileResource,
    ProductionPreviewBackgroundFileRemoveResource,
    ProductionTaskTypeResource,
    ProductionTaskTypeRemoveResource,
    ProductionTaskTypesResource,
    ProductionTaskStatusResource,
    ProductionTaskStatusRemoveResource,
    ProductionStatusAutomationResource,
    ProductionStatusAutomationRemoveResource,
    ProductionMetadataDescriptorResource,
    ProductionMetadataDescriptorsResource,
    ProductionMilestonesResource,
    ProductionScheduleItemsResource,
    ProductionTaskTypeScheduleItemsResource,
    ProductionAssetTypesScheduleItemsResource,
    ProductionEpisodesScheduleItemsResource,
    ProductionSequencesScheduleItemsResource,
    ProductionTimeSpentsResource,
)

routes = [
    ("/data/projects/open", OpenProjectsResource),
    ("/data/projects/all", AllProjectsResource),
    ("/data/projects/<project_id>/team", ProductionTeamResource),
    (
        "/data/projects/<project_id>/task-types",
        ProductionTaskTypesResource,
    ),
    (
        "/data/projects/<project_id>/team/<person_id>",
        ProductionTeamRemoveResource,
    ),
    (
        "/data/projects/<project_id>/settings/asset-types",
        ProductionAssetTypeResource,
    ),
    (
        "/data/projects/<project_id>/settings/asset-types/<asset_type_id>",
        ProductionAssetTypeRemoveResource,
    ),
    (
        "/data/projects/<project_id>/settings/task-types",
        ProductionTaskTypeResource,
    ),
    (
        "/data/projects/<project_id>/settings/task-types/<task_type_id>",
        ProductionTaskTypeRemoveResource,
    ),
    (
        "/data/projects/<project_id>/settings/task-status",
        ProductionTaskStatusResource,
    ),
    (
        "/data/projects/<project_id>/settings/task-status/<task_status_id>",
        ProductionTaskStatusRemoveResource,
    ),
    (
        "/data/projects/<project_id>/settings/status-automations",
        ProductionStatusAutomationResource,
    ),
    (
        "/data/projects/<project_id>/settings/status-automations/<status_automation_id>",
        ProductionStatusAutomationRemoveResource,
    ),
    (
        "/data/projects/<project_id>/settings/preview-background-files/<preview_background_file_id>",
        ProductionPreviewBackgroundFileRemoveResource,
    ),
    (
        "/data/projects/<project_id>/settings/preview-background-files",
        ProductionPreviewBackgroundFileResource,
    ),
    (
        "/data/projects/<project_id>/metadata-descriptors",
        ProductionMetadataDescriptorsResource,
    ),
    (
        "/data/projects/<project_id>/metadata-descriptors/<descriptor_id>",
        ProductionMetadataDescriptorResource,
    ),
    ("/data/projects/<project_id>/milestones", ProductionMilestonesResource),
    (
        "/data/projects/<project_id>/schedule-items",
        ProductionScheduleItemsResource,
    ),
    (
        "/data/projects/<project_id>/schedule-items/task-types",
        ProductionTaskTypeScheduleItemsResource,
    ),
    (
        "/data/projects/<project_id>/schedule-items/<task_type_id>/asset-types",
        ProductionAssetTypesScheduleItemsResource,
    ),
    (
        "/data/projects/<project_id>/schedule-items/<task_type_id>/episodes",
        ProductionEpisodesScheduleItemsResource,
    ),
    (
        "/data/projects/<project_id>/schedule-items/<task_type_id>/sequences",
        ProductionSequencesScheduleItemsResource,
    ),
    ("/data/projects/<project_id>/time-spents", ProductionTimeSpentsResource),
]

blueprint = Blueprint("projects", "projects")
api = configure_api_from_blueprint(blueprint, routes)



import psutil
import redis
import requests
import datetime

from flask import Response, abort
from flask_restful import Resource
from zou import __version__

from zou.app import app, config
from zou.app.utils import permissions, shell, date_helpers
from zou.app.services import projects_service, stats_service, persons_service

from flask_jwt_extended import jwt_required


class IndexResource(Resource):
    def get(self):
        """
        Get API name and version.
        ---
        tags:
          - Index
        responses:
            200:
                description: API name and version
        """
        return {"api": config.APP_NAME, "version": __version__}


class BaseStatusResource(Resource):
    def get_status(self):
        is_db_up = True
        try:
            projects_service.get_or_create_status("Open")
        except Exception:
            is_db_up = False

        is_kv_up = True
        try:
            store = redis.StrictRedis(
                host=config.KEY_VALUE_STORE["host"],
                port=config.KEY_VALUE_STORE["port"],
                db=config.AUTH_TOKEN_BLACKLIST_KV_INDEX,
                password=config.KEY_VALUE_STORE["password"],
                decode_responses=True,
            )
            store.get("test")
        except redis.ConnectionError:
            is_kv_up = False

        is_es_up = True
        try:
            requests.get(
                "http://{host}:{port}".format(
                    host=config.EVENT_STREAM_HOST,
                    port=config.EVENT_STREAM_PORT,
                )
            )
        except Exception:
            is_es_up = False

        is_jq_up = True
        try:
            host = config.KEY_VALUE_STORE["host"]
            port = config.KEY_VALUE_STORE["port"]
            db = config.KV_JOB_DB_INDEX
            password = config.KEY_VALUE_STORE["password"]
            if password:
                url = "redis://:%s@%s:%s/%s" % (password, host, port, db)
            else:
                url = "redis://%s:%s/%s" % (host, port, db)
            args = ["rq", "info", "--url", url]
            out = shell.run_command(args)
            is_jq_up = b"0 workers" not in out
        except Exception:
            app.logger.error("Job queue is not accessible", exc_info=1)
            is_jq_up = False

        is_indexer_up = True
        try:
            requests.get(
                "{protocol}://{host}:{port}".format(
                    protocol=config.INDEXER["protocol"],
                    host=config.INDEXER["host"],
                    port=config.INDEXER["port"],
                )
            )
        except Exception:
            is_indexer_up = False

        version = __version__

        return (
            config.APP_NAME,
            version,
            is_db_up,
            is_kv_up,
            is_es_up,
            is_jq_up,
            is_indexer_up,
        )


class StatusResource(BaseStatusResource):
    def get(self):
        """
        Retrieve API name, version and status.
        ---
        tags:
          - Index
        responses:
            200:
                description: API name, version and status
        """
        (
            api_name,
            version,
            is_db_up,
            is_kv_up,
            is_es_up,
            is_jq_up,
            is_indexer_up,
        ) = self.get_status()

        return {
            "name": api_name,
            "version": version,
            "database-up": is_db_up,
            "key-value-store-up": is_kv_up,
            "event-stream-up": is_es_up,
            "job-queue-up": is_jq_up,
            "indexer-up": is_indexer_up,
        }


class StatusResourcesResource(BaseStatusResource):
    def get(self):
        """
        Retrieve date and CPU, memory and jobs stats.
        ---
        tags:
          - Index
        responses:
            200:
                description: Date and CPU, memory and jobs stats
        """
        loadavg = list(psutil.getloadavg())

        cpu_stats = {
            "percent": psutil.cpu_percent(interval=1, percpu=True),
            "loadavg": {
                "last 1 min": loadavg[0],
                "last 5 min": loadavg[1],
                "last 10 min": loadavg[2],
            },
        }

        memory_stats = {
            "total": psutil.virtual_memory().total,
            "used": psutil.virtual_memory().used,
            "available": psutil.virtual_memory().available,
            "percent": psutil.virtual_memory().percent,
        }

        nb_jobs = 0
        if config.ENABLE_JOB_QUEUE:
            from zou.app.stores.queue_store import job_queue

            registry = job_queue.started_job_registry
            nb_jobs = registry.count
        job_stats = {
            "running_jobs": nb_jobs,
        }

        return {
            "date": datetime.now().isoformat(),
            "cpu": cpu_stats,
            "memory": memory_stats,
            "jobs": job_stats,
        }


class TxtStatusResource(BaseStatusResource):
    def get(self):
        """
        Retrieve API name, version and status as txt.
        ---
        tags:
          - Index
        responses:
            200:
                description: API name, version and status as txt
        """
        (
            api_name,
            version,
            is_db_up,
            is_kv_up,
            is_es_up,
            is_jq_up,
            is_indexer_up,
        ) = self.get_status()

        text = """name: %s
version: %s
database-up: %s
event-stream-up: %s
key-value-store-up: %s
job-queue-up: %s
indexer-up: %s
""" % (
            api_name,
            version,
            "up" if is_db_up else "down",
            "up" if is_kv_up else "down",
            "up" if is_es_up else "down",
            "up" if is_jq_up else "down",
            "up" if is_indexer_up else "down",
        )
        return Response(text, mimetype="text")


class InfluxStatusResource(BaseStatusResource):
    def get(self):
        """
        Retrieve status of database and time.
        ---
        tags:
          - Index
        responses:
            200:
                description: Status of database, key value, event stream, job queue and time
        """
        (
            _,
            _,
            is_db_up,
            is_kv_up,
            is_es_up,
            is_jq_up,
            is_indexer_up,
        ) = self.get_status()

        return {
            "database-up": int(is_db_up),
            "key-value-store-up": int(is_kv_up),
            "event-stream-up": int(is_es_up),
            "job-queue-up": int(is_jq_up),
            "indexer-up": int(is_indexer_up),
            "time": datetime.datetime.timestamp(
                date_helpers.get_utc_now_datetime()
            ),
        }


class StatsResource(Resource):
    @jwt_required()
    def get(self):
        """
        Retrieve main stats.
        ---
        tags:
          - Index
        responses:
            403:
                description: Permission denied
            200:
                description: Main stats
        """
        if not permissions.has_admin_permissions():
            abort(403)
        return stats_service.get_main_stats()


class ConfigResource(Resource):
    def get(self):
        """
        Get basic configuration for the current instance.
        ---
        tags:
          - Index
        responses:
            200:
                description: Configuration object including self-hosted status,
                    Crisp token, indexer configuration, SAML status, and dark
                    theme status.
        """
        organisation = persons_service.get_organisation()
        conf = {
            "is_self_hosted": config.IS_SELF_HOSTED,
            "crisp_token": config.CRISP_TOKEN,
            "dark_theme_by_default": organisation["dark_theme_by_default"],
            "indexer_configured": (
                len(config.INDEXER["key"]) > 0
                and config.INDEXER["key"] != "masterkey"
            ),
            "saml_enabled": config.SAML_ENABLED,
            "saml_idp_name": config.SAML_IDP_NAME,
            "default_locale": config.DEFAULT_LOCALE,
            "default_timezone": config.DEFAULT_TIMEZONE,
        }
        if config.SENTRY_KITSU_ENABLED:
            conf["sentry"] = {
                "dsn": config.SENTRY_KITSU_DSN,
                "sampleRate": config.SENTRY_KITSU_SR,
            }
        return conf


class TestEventsResource(Resource):
    def get(self):
        """
        Generate a main:test event.
        ---
        tags:
          - Index
        responses:
            200:
                description: Success flage
        """
        from zou.app.utils import events

        events.emit("main:test", data={}, persist=False, project_id=None)
        return {"success": True}



from flask import request, current_app
from flask_jwt_extended import jwt_required

from sqlalchemy.orm import aliased
from sqlalchemy.exc import IntegrityError

from zou.app.mixin import ArgsMixin
from zou.app.models.entity import Entity
from zou.app.models.person import Person
from zou.app.models.project import Project
from zou.app.models.task import Task

from zou.app.services import (
    user_service,
    tasks_service,
    deletion_service,
    entities_service,
    assets_service,
)
from zou.app.utils import permissions

from zou.app.services.exception import WrongTaskTypeForEntityException

from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource


class TasksResource(BaseModelsResource, ArgsMixin):
    def __init__(self):
        BaseModelsResource.__init__(self, Task)

    def check_read_permissions(self, options=None):
        return True

    def add_project_permission_filter(self, query):
        if permissions.has_vendor_permissions():
            query = query.filter(user_service.build_assignee_filter())
        elif not permissions.has_admin_permissions():
            query = query.join(Project).filter(
                user_service.build_related_projects_filter()
            )
        return query

    def build_filters(self, options):
        (
            many_join_filter,
            in_filter,
            name_filter,
            criterions,
        ) = super().build_filters(options)
        if "project_id" in criterions:
            del criterions["project_id"]
        if "episode_id" in criterions:
            del criterions["episode_id"]
        return (
            many_join_filter,
            in_filter,
            name_filter,
            criterions,
        )

    def apply_filters(self, query, options):
        query = super().apply_filters(query, options)

        project_id = options.get("project_id", None)
        episode_id = options.get("episode_id", None)
        if episode_id is not None:
            Sequence = aliased(Entity)
            query = (
                query.join(Entity, Task.entity_id == Entity.id)
                .join(Sequence, Entity.parent_id == Sequence.id)
                .filter(Sequence.parent_id == episode_id)
            )
        elif project_id is not None:
            query = query.join(Entity, Task.entity_id == Entity.id).filter(
                Entity.project_id == project_id
            )

        return query

    def post(self):
        """
        Create a task with data given in the request body. JSON format is
        expected. The model performs the validation automatically when
        instantiated.
        """
        try:
            data = request.json
            is_assignees = "assignees" in data
            assignees = None

            task_type = tasks_service.get_task_type(data["task_type_id"])
            entity = entities_service.get_entity(data["entity_id"])
            if task_type["for_entity"] == "Asset":
                if not assets_service.is_asset_dict(entity):
                    raise WrongTaskTypeForEntityException(
                        "Task type of the task does not match entity type."
                    )
            elif (
                entities_service.get_temporal_entity_type_by_name(
                    task_type["for_entity"]
                )["id"]
                != entity["entity_type_id"]
            ):
                raise WrongTaskTypeForEntityException(
                    "Task type of the task does not match entity type."
                )

            if is_assignees:
                assignees = data["assignees"]
                persons = Person.query.filter(Person.id.in_(assignees)).all()
                del data["assignees"]

            instance = self.model(**data)
            if assignees is not None:
                instance.assignees = persons
            instance.save()
            self.emit_create_event(instance.serialize())

            return (
                tasks_service.get_task(str(instance.id), relations=True),
                201,
            )

        except TypeError as exception:
            current_app.logger.error(str(exception), exc_info=1)
            return {"message": str(exception)}, 400

        except IntegrityError as exception:
            current_app.logger.error(str(exception), exc_info=1)
            return {"message": "Task already exists."}, 400


class TaskResource(BaseModelResource, ArgsMixin):
    def __init__(self):
        BaseModelResource.__init__(self, Task)

    def check_read_permissions(self, task):
        user_service.check_project_access(task["project_id"])
        user_service.check_entity_access(task["entity_id"])

    def check_update_permissions(self, task, data):
        user_service.check_supervisor_task_access(task, data)

    def check_delete_permissions(self, task):
        user_service.check_manager_project_access(task["project_id"])

    def post_update(self, instance_dict, data):
        tasks_service.clear_task_cache(instance_dict["id"])
        return instance_dict

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete a model corresponding at given ID and return it as a JSON
        object.
        """
        force = self.get_force()

        instance = self.get_model_or_404(instance_id)

        try:
            instance_dict = instance.serialize()
            self.check_delete_permissions(instance_dict)
            deletion_service.remove_task(instance_id, force=force)
            tasks_service.clear_task_cache(instance_id)
            self.post_delete(instance_dict)

        except IntegrityError as exception:
            current_app.logger.error(str(exception), exc_info=1)
            return {"message": str(exception)}, 400

        return "", 204
